package EFS::CLI::EFSServer;

use strict;
use warnings;
use English;

use base q{EFS::CLI};
use base q{EFS::DBI};

use EFS::FQDN qw( fqdn );

use Readonly;
use Params::Validate qw( :all );
use Date::Format;
use File::Slurp;
use File::Basename qw(basename dirname);
use Text::Password::Pronounceable;
use Try::Tiny;

Readonly my $SCREEN_WIDTH => 1024;

# XXX: These are oracle-specific, but seem to work fine for MySQL, too.
Readonly my $PASSWORD_LENGTH_MIN => 8;
Readonly my $PASSWORD_LENGTH_MAX => 30;

#
# XXX: This really should run against all the EFSservers, and display the
# results.
#
sub showcreds {
    my $class = shift;
    system q{klist};
    return 1;
}

sub create {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        name     => { type => SCALAR },
        hosttype => { type => SCALAR },
    });

    

}

sub define {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        name     => { type => SCALAR },
        hosttype => { type => SCALAR },
        clone    => { type => SCALAR, optional => 1 },
        cells    => { type => SCALAR, optional => 1 },
    });

    my $resultset = $class->get_resultset;

    my ($efsserver) = $resultset->search( { name => $args{name} } );

    if ( $efsserver ) {
        if ( $efsserver->hosttype eq $args{hosttype} ) {
            STDOUT->print(
                qq{Verified existing definition of efsserver '$args{name}'\n},
            );
        } else {
            $efsserver->set_column( hosttype => $args{hosttype} );
            $efsserver->update;
            STDOUT->print(
                qq{Set hosttype of efsserver '$args{name}' },
                qq{to '$args{hosttype}'\n}
            );
        }
    } else {
        $efsserver = $resultset->create({
            name     => $args{name},
            hosttype => $args{hosttype},
        });
        STDOUT->print(
            qq{Defined efsserver '$args{name}' in with },
            qq{hosttype '$args{hosttype}'\n}
        );
    }

    return $efsserver;

}

sub initssl {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        name      => { type => SCALAR, optional => 1 },
        hosttype  => { type => SCALAR, optional => 1 },
        bootstrap => { type => SCALAR, optional => 1 },
    });

    my $sslserver_rs = $class->get_resultset( q{SSLServer} );
    my ($primary) = $sslserver_rs->search( { type => q{primary} } );

    if ( not $primary ) {
        die qq{No primary SSLServer has been defined yet.\n};
    }

    if ( not $primary->efsserver->is_localhost ) {
        die(
            q{_EFS_REDIRECT_SERVER: } . $primary->efsserver->name . qq{\n},
        );
    }

    if ( $args{name} && $args{hosttype} ) {
        die qq{Mutually exclusive arguments: name and hosttype\n};
    }

    my @efsservers = ();

    my $resultset = $class->get_resultset;

    if ( $args{hosttype} ) {
        @efsservers = $resultset->search({ hosttype => $args{hosttype} });
    } elsif ( $args{name} ) {
        @efsservers = $resultset->search({ name => $args{name} });
    } else {
        @efsservers = $resultset->all;
    }

    # NOTE: This uses parallel of 1, because initssl_efsserver is effectively
    # single threaded anyway by running on the primary sslserver, and locking
    # /var/efs/ca.

    my $fm_efsservers = EFS::ForkManager->new({
        list     => \@efsservers,
        parallel => 1,
        action   => sub { return EFS::CLI::SSLServer->initssl_efsserver(shift) },

    });

    $fm_efsservers->run || die(
        qq{Unable to initssl on these efsservers:\n\t},
        join(qq{\n\t}, $fm_efsservers->get_failed) , qq{\n}
    );

    STDOUT->print( qq{Initialized SSL on all EFSServers\n} );

    return 1;

}

sub purgecerts {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        name      => { type => SCALAR, optional => 1 },
        hosttype  => { type => SCALAR, optional => 1 },
        bootstrap => { type => SCALAR, optional => 1 },
    });

    my $resultset = $class->get_resultset;

    my @efsservers = ();

    if ( $args{name} ) {
        @efsservers = $resultset->retrieve_bynames_assert( name => $args{name} );
    } elsif ( $args{hosttype} ) {
        @efsservers = $resultset->search( { hosttype => $args{hosttype} } );
    } else {
        @efsservers = $resultset->all;
    }

    my $sslserver_rs = $class->get_resultset(q{SSLServer});
    my ($sslserver) = $sslserver_rs->search( { type => q{primary} } );

    EFS::RPC->execute_on(
        efsservers => [ $sslserver->efsserver ],
        class      => q{EFS::CLI::SSLServer},
        method     => q{purgecerts_sslserver},
        args       => [ map { $_->name => 1 } @efsservers ],
    );

    my $fm_efsservers = EFS::ForkManager->new({
        list   => \@efsservers,
        action => sub {
            my $efsserver = shift;
            EFS::RPC->execute_on(
                efsservers => [$efsserver],
                class      => $class,
                method     => q{purgecerts_efsserver},
                args       => [%args],
            );
        },
    });

    $fm_efsservers->run || die(
        qq{Unable to purge old certificates on:\n\t},
        join ( qq{\n\t}, $fm_efsservers->get_failed ), qq{\n}
    );

    return 1;

}

sub purgecerts_efsserver {

    my $class = shift;
    my %args  = @_;

    my $name = fqdn;

    my $dirlock = EFS::DirLock->new({
        directory => q{/var/efs},
        retry     => 1,
        command   => qq{$class->purgecerts_efsserver},
    });

    my $sslroot = q{/var/efs/ssl};

    my $sslcurrent = readlink ($sslroot) || die qq{Unable to readlink $sslroot: $ERRNO\n};

    my $dir = IO::Dir->new(q{/var/efs}) || die qq{Unable to opendir /var/efs: $ERRNO\n};

    foreach my $dirent ( $dir->read ) {
        next if not $dirent =~ m{\Assl-\d{8}-\d{6}\z}ms;
        next if $dirent eq $sslcurrent;
        EFS::Util->RemoveDirHierarchy( top => qq{/var/efs/$dirent} );
    }

    STDOUT->print( qq{Purged all old SSL certificates on $name\n} );

    return 1;

}

sub newcerts {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        name      => { type => SCALAR, optional => 1 },
        hosttype  => { type => SCALAR, optional => 1 },
        bootstrap => { type => SCALAR, optional => 1 },
    });

    my $resultset = $class->get_resultset;

    my @efsservers = ();

    if ( $args{name} ) {
        @efsservers = $resultset->retrieve_bynames_assert( name => $args{name} );
    } elsif ( $args{hosttype} ) {
        @efsservers = $resultset->search( { hosttype => $args{hosttype} } );
    } else {
        @efsservers = $resultset->all;
    }

    #
    # Update the localhost's SSL certs before attempting to update any others,
    # since we need the SSL certs to be stable locally before attempting to
    # use them to run methods via EFS::RPC.
    #
    if ( grep { $_->is_localhost } @efsservers ) {
        $class->newcerts_efsserver( %args );
        @efsservers = grep { not $_->is_localhost } @efsservers;
    }

    my $fm_efsservers = EFS::ForkManager->new({
        list   => \@efsservers,
        action => sub {
            my $efsserver = shift;
            EFS::RPC->execute_on(
                efsservers => [$efsserver],
                class      => $class,
                method     => q{newcerts_efsserver},
                args       => [%args],
            );
        },
    });

    $fm_efsservers->run || die(
        qq{Unable to install new certificates on:\n\t},
        join ( qq{\n\t}, $fm_efsservers->get_failed ), qq{\n}
    );

    return 1;

}

sub newcerts_efsserver {

    my $class = shift;
    my (%args) = @_;

    my $name = fqdn;

    my $dirlock = EFS::DirLock->new({
        directory => q{/var/efs},
        retry     => 1,
        command   => qq{$class->newcerts_efsserver},
    });

    my $sslroot = q{/var/efs/ssl};
    my $sslnew = time2str( qq{$sslroot-%Y%m%d-%H%M%S}, time, q{GMT} );

    if ( not -l $sslroot or not -f q{/var/efs/etc/openssl.conf} ) {
        die(qq{SSL has not yet been initialized on $name\n});
    }

    EFS::Util->MakeDir( path => $sslnew, mode => 0700 );

    # Create a new certificate request

    EFS::CLI::SSLServer->create_passphrase( dir => $sslnew );

    $class->create_certificate_request(
        config  => q{/var/efs/etc/openssl.conf},
        keyout  => qq{$sslnew/$name-key.pem},
        out     => qq{$sslnew/$name-req.pem},
        passout => qq{file:$sslnew/passphrase},
    );

    # Send it to the primary SSLServer

    my $sslserver_rs = $class->get_resultset(q{SSLServer});
    my ($sslserver)  = $sslserver_rs->search({ type => q{primary} });
    my $efsserver    = $sslserver->efsserver;

    EFS::Rsync->new({
        dest_efsserver => $efsserver,
        src_path       => qq{$sslnew/$name-req.pem},
        dest_path      => qq{/var/efs/ca/reqs/$name-req.pem},
    });

    # Issue the certificate on the primary SSLServer

    EFS::RPC->execute_on(
        efsservers => [$efsserver],
        class      => q{EFS::CLI::SSLServer},
        method     => q{create_server_certificate},
        args       => [
            config => q{/var/efs/ca/openssl.conf},
            in     => qq{/var/efs/ca/reqs/$name-req.pem},
            out    => qq{/var/efs/ca/certs/$name-cert.pem},
        ],
    );

    EFS::RPC->execute_on(
        efsservers => [$efsserver],
        class      => q{EFS::Rsync},
        method     => q{new},
        args       => [{
            dest_efsserver_name => $name,
            src_path            => qq{/var/efs/ca/certs/$name-cert.pem},
            dest_path           => qq{$sslnew/$name-cert.pem},
        }],
    );

    # Install the new ssl directory

    EFS::Util->MakeLink(
        from => $sslroot,
        to   => basename($sslnew),
    );

    STDOUT->print(qq{Installed new certificates on $name\n});

    return 1;

}

sub newkeys {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        bootstrap => { type => SCALAR, optional => 1 },
    });

    if ( not $args{bootstrap} ) {
        my $project_rs = $class->get_resultset(q{Project});
        my $project = $project_rs->retrieve_bynames_assert(
            metaproj => q{efs},
            name     => q{config-efsserver},
        );
        return EFS::RPC->execute(
            cell   => $project->cell,
            class  => $class,
            method => q{newkeys_rpc},
            args   => [%args],
        );
    } else {
        return $class->newkeys_rpc(%args);
    }

}

sub newkeys_rpc {

    my $class = shift;
    my %args  = @_;

    my $project_lock;

    if ( not $args{bootstrap} ) {

        my $history_rs = $class->get_resultset(q{History});
        my $history = $history_rs->retrieve_assert( id => $args{history} );

        my $project_rs = $class->get_resultset(q{Project});

        my $project = $project_rs->retrieve_bynames_assert(
            metaproj => q{efs},
            name     => q{config-efsserver},
        );

        my $project_dir = $project->get_path( q{dev} );

        $project_lock = EFS::DirLock->new({
            directory => $project_dir,
            command   => $history->request,
        });

    }

    # Create a new ssh key pair, in /var/efs/.ssh/id_rsa_efsd-YYYYMMDD-HHMMSS

    EFS::Util->MakeDir( path => q{/var/efs/.ssh}, mode => 0700 );

    my $keylink = q{/var/efs/.ssh/efsd};
    my $keyfile = time2str( qq{$keylink-%Y%m%d-%H%M%S}, time, q{GMT} );

    my $errors  = qx{ ssh-keygen -t rsa -f $keyfile -N '' < /dev/null 2>&1 };
    $CHILD_ERROR and die qq{Unable to generate new ssh key:\n$errors};

    my $public = read_file(qq{$keyfile.pub}) ||
        die qq{Unable to read $keyfile.pub: $ERRNO\n};
    my $private = read_file($keyfile) ||
        die qq{Unable to read $keyfile: $ERRNO\n};

    my $resultset = $class->get_resultset;

    my @efsservers = $resultset->all;

    # Add the public key to all efsserver's /root/.ssh/ authorized_keys files

    my $fm_newpubkey = EFS::ForkManager->new({
        list => \@efsservers,
        action => sub {
            my $efsserver = shift;
            if ( not $efsserver->is_localhost ) {
                EFS::RPC->execute_on(
                    efsservers => [$efsserver],
                    class      => $class,
                    method     => q{install_keys},
                    args       => [
                        name    => basename($keyfile),
                        public  => $public,
                        private => $private,
                    ],
                );
            }
            EFS::RPC->execute_on(
                efsservers => [$efsserver],
                class      => $class,
                method     => q{sync_public_keys},
            );
            return 1;
        },
    });

    $fm_newpubkey->run || die(
        qq{Unable to distribute and authorize the new public key on:\n\t},
        join ( qq{\n\t}, $fm_newpubkey->get_failed ), qq{\n},
    );

    # Change the  /var/efs/.ssh/efsd symlink globally

    my $fm_newprivkey = EFS::ForkManager->new({
        list => \@efsservers,
        action => sub {
            my $efsserver = shift;
            EFS::RPC->execute_on(
                efsservers => [$efsserver],
                class      => q{EFS::Util},
                method     => q{MakeLink},
                args       => [
                    from      => $keylink,
                    from_show => qq{$efsserver:$keylink},
                    to        => basename($keyfile),
                ],
            );
        },
    });

    $fm_newprivkey->run || die(
        qq{Unable to configure the new key on:\n},
        join ( qq{\n\t}, $fm_newprivkey->get_failed ), qq{\n},
    );

    STDOUT->print(
        qq{Distributed new efsd ssh key to all efsservers globally.\n},
    );

    return 1;

}

sub newdbpw {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        bootstrap => { type => SCALAR, optional => 1 },
    });

    if ( not $args{bootstrap} ) {
        my $project_rs = $class->get_resultset(q{Project});
        my $project = $project_rs->retrieve_bynames_assert(
            metaproj => q{efs},
            name     => q{config-efsserver},
        );
        return EFS::RPC->execute(
            cell   => $project->cell,
            class  => $class,
            method => q{newdbpw_rpc},
            args   => [%args],
        );
    } else {
        return $class->newdbpw_rpc(%args);
    }

}

sub newdbpw_rpc {

    my $class = shift;
    my %args  = @_;

    my $project_lock;

    if ( not $args{bootstrap} ) {
        my $history_rs = $class->get_resultset(q{History});
        my $history = $history_rs->retrieve_assert( id => $args{history}, );
        my $project_rs = $class->get_resultset(q{Project});
        my $project = $project_rs->retrieve_bynames_assert(
            metaproj => q{efs},
            name     => q{config-efsserver},
        );
        my $project_dir = $project->get_path(q{dev});
        $project_lock = EFS::DirLock->new({
            directory => $project_dir,
            command   => $history->request,
        });
    }

    # Create a new DB password

    my $password = Text::Password::Pronounceable->generate(
        $PASSWORD_LENGTH_MIN,
        $PASSWORD_LENGTH_MAX,
    );

    my @efsservers = $class->get_resultset->all;

    # Distribute the new passwd file to all efsservers

    my $fm_newdbpw = EFS::ForkManager->new({
        list => \@efsservers,
        action => sub {
            my $efsserver = shift;
            EFS::RPC->execute_on(
                efsservers => [$efsserver],
                class      => $class,
                method     => q{install_dbpw},
                args       => [$password],
            );
            return 1;
        },
    });

    $fm_newdbpw->run || die(
        qq{Unable to distribute new database password on:\n\t},
        join ( qq{\n\t}, $fm_newdbpw->get_failed ), qq{\n},
    );

    # Change the password in the database

    my $config = EFS::Config::Server->new;

    my $rwuser = $config->get_value( q{database}, q{rwuser} );

    $class->set_db_password(
        user     => $rwuser,
        password => $password,
    );

    # Purge the old password everwhere.

    my $fm_olddbpw = EFS::ForkManager->new({
        list => \@efsservers,
        action => sub {
            my $efsserver = shift;
            EFS::RPC->execute_on(
                efsservers => [$efsserver],
                class      => $class,
                method     => q{purge_dbpw},
                args       => [ $password, ],
            );
            return 1;
        },
    });

    $fm_olddbpw->run || die(
        qq{Unable to purge old database password on:\n\t},
        join ( qq{\n\t}, $fm_olddbpw->get_failed ), qq{\n},
    );

    return 1;

}

sub install_dbpw {

    my $class    = shift;
    my $password = shift;

    my $fqdn = fqdn;

    my $config  = EFS::Config::Server->new;
    my $db_user = $config->get_value( q{database}, q{rwuser} );

    my $password_dir  = q{/var/efs/etc/passwd};
    my $password_file = qq{$password_dir/$db_user};

    my $io = IO::File->new(qq{>>$password_file}) ||
        die(qq{Unable to append to $password_file: $ERRNO\n});

    $io->print(qq{$password\n}) ||
        die qq{Unable to write to $password_file: $ERRNO\n};

    $io->close ||
        die qq{Unable to close $password_file: $ERRNO\n};

    STDOUT->print(qq{Installed new database password on $fqdn\n});

    return 1;

}

sub purge_dbpw {

    my $class    = shift;
    my $password = shift;

    my $fqdn = fqdn;

    my $config  = EFS::Config::Server->new;
    my $db_user = $config->get_value( q{database}, q{rwuser} );

    my $password_dir  = q{/var/efs/etc/passwd};
    my $password_file = qq{$password_dir/$db_user};
    my $new_file      = qq{$password_file.$PID};

    my $io = IO::File->new(qq{>$new_file}) ||
        die qq{Unable to write to $new_file: $ERRNO\n};

    $io->print(qq{$password\n}) ||
        die qq{Unable to write to $new_file: $ERRNO\n};

    $io->close ||
        die qq{Unable to close $new_file: $ERRNO\n};

    rename( $new_file, $password_file ) ||
        die qq{Unable to rename $new_file -> $password_file: $ERRNO\n};

    STDOUT->print( qq{Purged old database passwords on $fqdn\n} );

    return 1;

}

sub install_keys {

    my $class = shift;

    my $fqdn = fqdn;

    my (%args) = validate(  @_, {
        name    => { type => SCALAR },
        public  => { type => SCALAR },
        private => { type => SCALAR },
    });

    foreach my $file (qw( public private )) {

        my $name = qq{/var/efs/.ssh/$args{name}};

        if ( $file eq q{public} ) {
            $name .= q{.pub};
        }

        my $io = IO::File->new(qq{>$name}) ||
            die qq{Unable to open $name: $ERRNO\n};

        $io->print( $args{$file} ) ||
            die qq{Unable to write to $name: $ERRNO\n};

        $io->close ||
            die qq{Unable to close $name: $ERRNO\n};

        if ( $file eq q{private} ) {
            ##  no critic (MagicNumbers)
            chmod( 0600, $name ) ||
                die qq{Unable to chmod $name: $ERRNO\n};
        }

    }

    STDOUT->print( qq{Installed new EFS ssh key $args{name}\n} );

    return 1;

}

sub sync_public_keys {

    my $class = shift;

    my $fqdn = fqdn;

    # Collect all the public keys found in /var/efs/.ssh

    my %public = ();

    my $dir = IO::Dir->new(q{/var/efs/.ssh}) ||
        die qq{Unable to opendir /var/efs/.ssh: $ERRNO\n};

    foreach my $dirent ( $dir->read ) {
        next if not $dirent =~ m{\A (efsd - \H+) \. pub \z}msx;
        my $comment = $1 ? $1 : q{};
        my $file = IO::File->new(qq{/var/efs/.ssh/$dirent}) ||
            die qq{Unable to open /var/efs/.ssh/$dirent: $ERRNO\n};
        my $line = $file->getline;
        chomp ($line);
        my ($type,$key) = split( m/\h+/msx, $line );
        $public{$key} = {
            type    => $type,
            comment => $comment
        };
    }

    # Parse authorized_keys

    my $authorized_keys = $class->authorized_keys_file;

    my %authorized = ();
    my %other      = ();

    if ( -f $authorized_keys ) {

        my $file = IO::File->new($authorized_keys) ||
            die qq{Unable to open $authorized_keys: $ERRNO\n};

        while ( defined ( my $line = $file->getline ) ) {

            chomp ($line);

            my ($type,$key,$comment) = split( m/\h+/msx, $line );

            if ( $comment =~ m{\A efsd - \d{8} - \d{6} \z}msx ) {
                $authorized{$key} = {
                    type    => $type,
                    comment => $comment,
                };
            } else {
                $other{$key} = {
                    type    => $type,
                    comment => $comment,
                };
            }

        }

    }

    # What's missing?

    my %missing = ();

    foreach my $key ( keys %public ) {
        if ( not $authorized{$key} ) {
            $missing{$key} = $public{$key};
        }
    }

    # What's obsolete?

    my %obsolete = ();

    foreach my $key ( keys %authorized ) {
        if ( not $public{$key} ) {
            $obsolete{$key} = $authorized{$key};
        }
    }

    if ( not %missing and not %obsolete ) {
        STDOUT->print( qq{All efsd public keys are authorized on $fqdn\n} );
        return 1;
    }

    my $content = q{};

    foreach my $key ( keys %other ) {
        my $type    = $other{$key}->{type};
        my $comment = $other{$key}->{comment};
        $content    .= qq{$type $key $comment\n};
    }

    foreach my $key ( keys %public ) {
        my $type    = $public{$key}->{type};
        my $comment = $public{$key}->{comment};
        $content    .= qq{$type $key $comment\n};
    }

    my $dirname = dirname $authorized_keys;
    my $dirmode = $dirname =~ m{\.ssh$}msx ? 0700 : 0755;

    if ( not -d $dirname ) {
        EFS::Util->MakeDir( path => $dirname, mode => $dirmode );
    }

    EFS::Config->update_file(
        file    => $authorized_keys,
        content => $content,
    );

    STDOUT->print( qq{Authorized all efsd ssh keys on $fqdn\n} );

    return 1;

}

sub authorized_keys_file {

    my $class = shift;

    my $sshd_config_file = q{/etc/ssh/sshd_config};

    my $io = IO::File->new( $sshd_config_file ) ||
        die qq{Unable to open $sshd_config_file: $ERRNO\n};

    my %sshd_config = ();

    while ( defined( $_ = $io->getline ) ) {
        s{\#.*}{}gmsx;
        s{^\s+}{}gmsx;
        chomp;
        next if not $_;
        my ($key,$value) = split( m{\s+}msx, $_, 2 );
        $sshd_config{$key} = $value;
    }

    $io->close ||
        die qq{Unable to close $sshd_config_file: $ERRNO\n};

    my $authorized_keys = q{};

    if ( $sshd_config{ AuthorizedKeysFile } ) {
        $authorized_keys = $sshd_config{ AuthorizedKeysFile };
        $authorized_keys =~ s{%u}{root}gmsx;
    } else {
        $authorized_keys = $class->root_sshdir . q{/authorized_keys};
    }

    return $authorized_keys;

}

# NOTE: During the bootstrap process, these directories may not be present.
# We need a better mechanism for determining the default .ssh dir location,
# which is OS specific.
sub root_sshdir {
    my $class = shift;
    if ( -d q{/root} ) {
        return q{/root/.ssh};
    } else {
        return q{/.ssh};
    }
}

sub sshid {
    # XXX: We could make this configurable, but if the paths aren't
    # the same on ALL efs servers, that will complicate key
    # distribution significantly.
    return q{/var/efs/.ssh/efsd};
}

sub purgekeys {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
    });

    my $resultset = $class->get_resultset;

    my @efsservers = $resultset->all;

    my $fm_efsservers = EFS::ForkManager->new({
        list => \@efsservers,
        action => sub {
            return EFS::RPC->execute_on(
                efsservers => [shift],
                class      => $class,
                method     => q{purgekeys_efsserver},
            );
        },
    });

    $fm_efsservers->run || die(
        qq{Unable to purge old keys on:\n},
        join ( qq{\n\t}, $fm_efsservers->get_failed ), qq{\n},
    );

    STDOUT->print( qq{Distributed new efsd ssh key to all efsservers globally.\n} );

    return 1;

}

sub purgekeys_efsserver {

    my $class = shift;

    my $fqdn = fqdn;

    my $latest = readlink(q{/var/efs/.ssh/efsd}) ||
        die qq{Unable to readlink /var/efs/.ssh/efsd: $ERRNO\n};

    my $dir = IO::Dir->new(q{/var/efs/.ssh}) ||
        die qq{Unable to opendir /var/efs/.ssh: $ERRNO\n};

    foreach my $dirent ( $dir->read ) {
        if ( $dirent !~ m{\A (?:efsd - \S+) (?:\. pub)? \z}msx ) {
            next;
        }
        if ( $dirent eq $latest || $dirent eq qq{$latest.pub} ) {
            next;
        }
        unlink (qq{/var/efs/.ssh/$dirent}) ||
            die qq{Unable to unlink /var/efs/.ssh/$dirent: $ERRNO\n};
        STDOUT->print( qq{Removed old ssh key file /var/efs/.ssh/$dirent\n} );
    }

    return $class->sync_public_keys;

}

sub dumpmaps {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        name     => { type => SCALAR, optional => 1 },
        hosttype => { type => SCALAR, optional => 1 },
    });

    if ( $args{name} && $args{hosttype} ) {
        die qq{Mutually exclusive arguments: name and hosttype\n};
    }

    my $resultset = $class->get_resultset;

    my @efsservers = ();

    if ( $args{name} ) {
        @efsservers = ( $resultset->retrieve_bynames_assert( name => $args{name} ) );
    } elsif ( $args{hosttype} ) {
        @efsservers = $resultset->search({ hosttype => $args{hosttype} });        
    } else {
        @efsservers = $resultset->all;
    }

    my $fm_efsservers = EFS::ForkManager->new({
        list => \@efsservers,
        action => sub {
            return EFS::RPC->execute_on(
                efsservers => [shift],
                class      => $class,
                method     => q{dumpmaps_efsserver},
            );
        },
    });

    $fm_efsservers->run || die(
        qq{Unable to dump auto maps on:\n},
        join ( qq{\n\t}, $fm_efsservers->get_failed ), qq{\n},
    );

    STDOUT->print( qq{Updated local auto maps on all efsservers.\n} );
    return 1;

}

sub dumpmaps_efsserver {

    my $class = shift;

    my $config = EFS::Config::Server->new;

    my $type = $config->get_value( q{server}, q{automounter} ) || q{autofs};

    my $install_dir = qq{/var/efs/etc/automaps/$type};

    EFS::Util->MakeDirHierarchy(
        top    => q{/var/efs},
        bottom => $install_dir,
    );

    my $dirlock = EFS::DirLock->new({
        directory => $install_dir,
        retry     => 1,
        command   => q{dumpmaps_efsserver},
    });

    my $resultset = $class->get_resultset;
    my $efsserver  = $resultset->retrieve_bynames_assert( name => fqdn );

    foreach my $cell ( $efsserver->cells ) {
        if ( $type eq q{autofs} ) {
            $cell->_generate_cell_auto_map_global( $install_dir );
            $cell->_generate_cell_auto_map_admin( $install_dir );
        } else {
            $cell->_generate_cell_amd_map_global( $install_dir );
            $cell->_generate_cell_amd_map_admin( $install_dir );
        }
    }

    STDOUT->print( qq{Updated locally installed $type maps\n} );
    return 1;

}

sub status {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        name     => { type => SCALAR, optional => 1 },
        hosttype => { type => SCALAR, optional => 1 },
    });

    my $resultset = $class->get_resultset;

    my @efsservers = ();

    if ( $args{name} ) {
        $efsservers[0] = $resultset->retrieve_bynames_assert( name => $args{name} );
    } elsif ( $args{hosttype} ) {
        @efsservers =
            grep { $_->exposed_cells }
                $resultset->search({ hosttype => $args{hosttype} });
    } else {
        @efsservers = grep { $_->exposed_cells } $resultset->all;
    }

    if ( not @efsservers ) {
        die qq{No EFS servers match the specified criteria\n};
    }

    my $fm_efsservers = EFS::ForkManager->new({
        list => \@efsservers,
        action => sub {
            my $efsserver = shift;
            return EFS::RPC->execute_on(
                efsservers => [$efsserver],
                class      => $class,
                method     => q{status_efsserver},
                args       => [ $efsserver->name ],
            );
        },
    });

    $fm_efsservers->run || die(
        qq{Unable to get status of the following efsservers:\n\t},
        join ( qq{\n\t}, $fm_efsservers->get_failed ), qq{\n},
    );

    return 1;

}

sub status_efsserver {

    my $class          = shift;
    my $efsserver_name = shift;

    my $efsserver_rs = $class->get_resultset(q{EFSServer});
    my $efsserver
        = $efsserver_rs->retrieve_bynames_assert( name => $efsserver_name );

    my $hosttype = $efsserver->hosttype;
    my $cells = join q{,}, map { $_->shortname } $efsserver->cells;

    my $history_rs = $class->get_resultset(q{History});
    my @history = $history_rs->search({
        efshost => $efsserver_name,
        status  => undef,
    });

    if ( not @history ) {
        STDOUT->print(
            qq{No active EFS commands on $efsserver_name ($hosttype, $cells)\n},
        );
        return 1;
    }

    my $ptable     = EFS::Proc::ProcessTable->new;
    my %process_of = ();

    foreach my $process ( @{ $ptable->table } ) {
        $process_of{ $process->pid } = $process;
    }

    my $table = Text::FormatTable->new(q{r  l  r  r  r  l  l});
    $table->head(
        q{History ID}, q{StartTime}, q{PID}, q{Size},
        q{RSS},        q{%CPU},      q{Request},
    );
    $table->rule;

    my $active = 0;

    foreach my $history (@history) {

        my $process = $process_of{ $history->pid };

        my $history_id = $history->id;

        if ( not $process ) {
            my ($refresh) = $history_rs->search({ id => $history_id });
            if ( not defined $refresh->status ) {
                $refresh->set_column( status => 0 );
                $refresh->update;
                STDOUT->print(
                    q{NOTE: Marked missing history process },
                    qq{(id = $history_id) on $efsserver_name as failed\n},
                );
            }
        } else {
            $table->row(
                $history_id,
                $history->starttime,
                $history->pid,
                EFS::DBI->size_shorthand( $process->size ),
                EFS::DBI->size_shorthand( $process->rss ),
                $process->pctcpu,
                $history->request
            );
            $active++;
        }

    }

    if ( not $active ) {
        STDOUT->print(
            qq{No active EFS commands on $efsserver_name ($hosttype, $cells)\n},
        );
    } else {
        STDOUT->print(
            qq{Active EFS commands on $efsserver_name ($hosttype, $cells)\n},
            $table->render($SCREEN_WIDTH),
        );
    }

    return 1;

}

sub dumpconfig {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        %EFS::CLI::validate_dumpconfig_args,
    });

    my $project_rs = $class->get_resultset(q{Project});

    my $project = $project_rs->retrieve_bynames_assert(
        metaproj => q{efs},
        name     => q{config-efsserver},
    );

    return EFS::RPC->execute(
        cell   => $project->cell,
        class  => $class,
        method => q{dumpconfig_rpc},
        args   => [%args],
    );

}

sub dumpconfig_rpc {

    my $class = shift;

    my %args = @_;

    my $history_rs = $class->get_resultset(q{History});
    my $history = $history_rs->retrieve_assert( id => $args{history} );

    my $project_rs = $class->get_resultset(q{Project});
    my $project = $project_rs->retrieve_bynames_assert(
        metaproj => q{efs},
        name     => q{config-efsserver},
    );

    my $release = EFS::CLI::AutoRelease->create_rpc(
        history  => $history->id,
        metaproj => q{efs},
        project  => q{config-efsserver},
    );

    my ($install) = $release->installs({ name => q{common} });

    my $install_dir = $install->get_path( q{dev} );

    EFS::Util->MakeDir( path => qq{$install_dir/etc} );

    my $resultset = $class->get_resultset;

    my @efsserver = sort { $a->name cmp $b->name } $resultset->all;

    my $output = <<"END_OF_HEADER";
#
# This file is auto-generated, so manual changes WILL be lost
#
END_OF_HEADER

    my $table = Text::FormatTable->new(q{l  r  l});

    foreach my $efsserver (@efsserver) {
        foreach my $cell ( $efsserver->cells ) {
            $table->row( $cell->name, $efsserver->hosttype,
                $efsserver->name );
        }
    }

    my $file = qq{$install_dir/etc/efsserver.txt};

    EFS::Config->update_file(
        file    => $file,
        content => $table->render($SCREEN_WIDTH),
    );

    EFS::CLI::AutoRelease->dist_rpc(
        history  => $history->id,
        metaproj => q{efs},
        project  => q{config-efsserver},
    );

    STDOUT->print(qq{Updated EFSServer config file\n});

    return 1;

}

sub checkaccess {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        name     => { type => SCALAR, optional => 1 },
        hosttype => { type => SCALAR, optional => 1 },
    });

    if ( $args{name} && $args{hosttype} ) {
        die(qq{Mutually exclusive arguments: name and hosttype\n});
    }

    my $resultset = $class->get_resultset;

    my @efsservers = ();

    if ( $args{name} ) {
        $efsservers[0]
            = $resultset->retrieve_bynames_assert( name => $args{name} );
    } elsif ( $args{hosttype} ) {
        @efsservers = $resultset->search({ hosttype => $args{hosttype} });
    } else {
        @efsservers = $resultset->all;
    }

    my $fm_efsserver = EFS::ForkManager->new({
        list => \@efsservers,
        action => sub {
            my ($efsserver) = @_;
            EFS::RPC->execute_on(
                efsservers => [$efsserver],
                class      => $class,
                method     => q{checkaccess_efsserver},
                args       => [ $efsserver->name ],
            );
        },
    });

    $fm_efsserver->run || die(qq{Access checks failed\n});

    return 1;

}

sub checkaccess_efsserver {

    my $class          = shift;
    my $efsserver_name = shift;

    my $efsserver_rs = $class->get_resultset(q{EFSServer});
    my $efsserver
        = $efsserver_rs->retrieve_bynames_assert( name => $efsserver_name );

    my $errors = 0;

    foreach my $cell ( $efsserver->cells ) {

        my @types = $cell->celltype eq q{dev} ? qw( dev dist ) : qw( dist );

        foreach my $type (@types) {

            my $admin_path = $cell->admin_path( type => $type );

            next if not $admin_path;    # It whines...

            my $tmpfile;

            try {
                $tmpfile = File::Temp->new( DIR => $admin_path );
            }
            catch {
                warn(
                    qq{Problems with access from $efsserver },
                    qq{to $type in $cell:\n}, $_,
                );
                $errors++;
            };

        }

    }

    if ($errors) {
        return 0;
    } else {
        STDOUT->print( qq{No access issues on $efsserver\n} );
        return 1;
    }

}

sub display {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        name     => { type => SCALAR, optional => 1 },
        hosttype => { type => SCALAR, optional => 1 },
    });

    if ( $args{name} && $args{hosttype} ) {
        die qq{Mutually exclusive arguments: name and hosttype\n};
    }

    my $resultset = $class->get_resultset;

    if ( $args{hosttype} ) {
        my @efsservers = $resultset->search({ hosttype => $args{hosttype} });
        $class->display_objects( objects => \@efsservers );
        return 1;
    } elsif ( not $args{name} ) {
        $class->display_objects( objects => [ $resultset->all ] );
        return 1;
    }

    my $efsserver = $resultset->retrieve_bynames_assert( name => $args{name} );

    $class->display_objects( objects => [$efsserver] );

    if ( my @attrs = $efsserver->attrs ) {
        $attrs[0]->display_objects( objects => \@attrs );
    }

    if ( my @cells = $efsserver->cells ) {
        $cells[0]->display_objects( objects => \@cells );
    }

    return 1;

}

sub restart {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        name     => { type => SCALAR, optional => 1 },
        hosttype => { type => SCALAR, optional => 1 },
    });

    if ( not $args{name} and not $args{hosttype} ) {
        die qq{At least one of -name or -hosttype must be specified.\n};
    }

    my %search = ();

    foreach my $key (qw( name hosttype )) {
        if ( $args{$key} ) {
            $search{$key} = $args{$key};
        }
    }

    my $resultset = $class->get_resultset;

    my @efsservers = $resultset->search( \%search );

    if ( not @efsservers ) {
        die qq{No EFS servers match the specified criteria\n};
    }

    my $sshid = $class->sshid;

    my $fm_efsservers = EFS::ForkManager->new({

        list => \@efsservers,

        trapoutput => 0,

        action => sub {

            my $efsserver = shift;

            STDOUT->print(qq{Restarting services on $efsserver\n});

            my $command = q{};

            if ( $efsserver->is_localhost ) {
                $command = q{/etc/init.d/efsserver restart 2>&1};
            } else {
                $command
                    = qq{ssh -i $sshid }
                        . q{-o StrictHostKeyChecking=no -o LogLevel=ERROR }
                            . qq{$efsserver /etc/init.d/efsserver }
                                . q{restart};
            }

            my $output = qx{ $command 2>&1 };

            if ( $CHILD_ERROR ) {
                die( qq{Unable to restart services on $efsserver\n}, $output );
            }

            return 1;
        }

    });

    $fm_efsservers->run || die(
        qq{Unable to restart services on the following efsservers:\n\t},
        join ( qq{\n\t}, $fm_efsservers->get_failed ), qq{\n}
    );

    return 1;

}

sub archivelogs {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        name     => { type => SCALAR, optional => 1 },
        hosttype => { type => SCALAR, optional => 1 },
    });

    my @efsservers = ();

    my %search = ();

    foreach my $key (qw( name hosttype )) {
        if ( $args{$key} ) {
            $search{$key} = $args{$key};
        }
    }

    my $resultset = $class->get_resultset;

    if (%search) {
        @efsservers = $resultset->search( \%search );
    } else {
        @efsservers = $resultset->all;
    }

    if ( not @efsservers ) {
        die qq{No EFS servers match the specified criteria\n};
    }

    my $fm_efsservers = EFS::ForkManager->new({

        list => \@efsservers,

        parallel => $#efsservers + 1,

        action => sub {

            my $efsserver = shift;

            return EFS::RPC->execute_on(
                efsservers => [$efsserver],
                class      => $class,
                method     => q{archivelogs_rpc},
                args       => [
                    history => $args{history},
                    name    => $efsserver->name,
                ],
            );

        },

    });

    $fm_efsservers->run || die(
        qq{Unable to archive logfiles on the following efsservers:\n\t},
        join ( qq{\n\t}, $fm_efsservers->get_failed ), qq{\n}
    );

    return 1;

}

sub archivelogs_rpc {

    my $class          = shift;
    my (%args)         = @_;

    my $efsserver_name = $args{name};

    my $efsserver_rs = $class->get_resultset(q{EFSServer});
    my $efsserver
        = $efsserver_rs->retrieve_bynames_assert( name => $efsserver_name );

    if ( not -d $EFS::logroot ) {
        STDOUT->print( qq{No logfiles to archive (yet) on $efsserver_name\n} );
        return 1;
    }

    STDOUT->print( qq{Archiving logfiles on $efsserver_name\n}, );

    my $metaproj_rs = EFS::DBI->get_resultset( q{Metaproj} );
    my $metaproj = $metaproj_rs->retrieve_bynames_assert( name => q{efs} );

    my $config = EFS::Config::Client->new;

    my $cell_rs = EFS::DBI->get_resultset( q{Cell} );
    my $cell = $cell_rs->retrieve_bynames_assert(
        name => $config->get_value( q{client}, q{cell} ),
    );

    my $logs = q{logs-} . $cell->shortname;

    my $project_rs = EFS::DBI->get_resultset( q{Project} );
    my ($project) = $project_rs->search({
        metaproj => $metaproj->id,
        name     => $logs,
    });

    if ( not $project ) {
        $project = EFS::CLI::Project->create(
            history  => $args{history},
            metaproj => q{efs},
            name     => $logs,
            cell     => $cell->name,
        );
    }

    my $release_rs = EFS::DBI->get_resultset( q{Release} );

    my ($release) = $release_rs->search({
        project => $project->id,
        name    => q{incr},
    });

    if ( not $release ) {
        $release = EFS::CLI::Release->create(
            history  => $args{history},
            metaproj => q{efs},
            project  => $logs,
            name     => q{incr},
        );
    }

    my @logfiles = qx{ find $EFS::logroot -type f -print };
    if ($CHILD_ERROR) {
        die qq{Unable to check $EFS::logroot for files\n};
    }

    if ( not @logfiles ) {
        STDOUT->print(
            qq{$EFS::logroot on $efsserver_name does not contain any logfiles.\n},
            qq{Nothing to archive\n}
        );
        return 1;
    }

    chomp (@logfiles);

    foreach my $logfile (@logfiles) {
        EFS::Logfile->archive(
            path => $logfile,
            verbose => 1,
            force => 1,
        );
    }

    return 1;

}

sub undef {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        name => { type => SCALAR },
    });

    my $resultset = $class->get_resultset;

    my $efsserver = $resultset->retrieve_bynames_assert( name => $args{name} );

    $efsserver->efsserver_cell->delete_all;
    $efsserver->attrs->delete_all;
    $efsserver->delete;

    STDOUT->print( qq{Undefined efsserver '$args{name}'\n} );

    return 1;

}

sub history {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        cell => { type => SCALAR },
        name => { type => SCALAR },
        %EFS::CLI::History::validation_specs,
    });

    $args{efsserver} = delete $args{name};

    return EFS::CLI::History->history(%args);

}

sub create_certificate_request {

    my $class = shift;
    my (%args) = @_;

    my $tmp = $args{keyout};
    $tmp =~ s{\.pem}{-tmp.pem}msx;

    my $openssl_cmd
        = (   qq{openssl req -config $args{config} }
            . q{-newkey rsa:2048 }
            . qq{-keyout $tmp -keyform PEM }
            . qq{-out $args{out} -outform PEM }
            . qq{-passout $args{passout} }
            . q{2>&1} );
    my $openssl_out = qx{ $openssl_cmd };

    if ($CHILD_ERROR) {
        die(qq{Unable to create new certificate request:\n$openssl_out});
    }

    $openssl_cmd
        = (   q{openssl rsa }
            . qq{-in $tmp -out $args{keyout} }
            . qq{-passin $args{passout} }
            . q{2>&1} );
    $openssl_out = qx{ $openssl_cmd };

    if ($CHILD_ERROR) {
        die(qq{Unable to remove passphrase from request:\n$openssl_out});
    }

    unlink ($tmp) || die qq{Unable to unlink $tmp: $ERRNO\n};

    STDOUT->print( qq{Created certificate request: $args{out}\n} );

    return 1;

}

1;

# Local Variables:
#   mode: cperl
#   cperl-indent-level: 4
#   fill-column: 78
# End:
# vim: expandtab shiftwidth=4:
