package EFS::CLI::Fileserver;

use strict;
use warnings;
use English;

use base q{EFS::CLI};
use base q{EFS::DBI};

use Readonly;
use Params::Validate qw( :all );

Readonly my $SCREEN_WIDTH => 1024;

sub define {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        cell => { type => SCALAR },
        name => { type => SCALAR },
    });

    my $cell_rs = $class->get_resultset( q{Cell} );
    my $cell = $cell_rs->retrieve_bynames_assert( name => $args{cell} );

    my $resultset = $class->get_resultset;

    my ($fileserver) = $resultset->search({
        cell => $cell->id,
        name => $args{name},
    });

    if ( $fileserver ) {
        STDOUT->print( qq{Fileserver '$fileserver' in cell '$cell' already exists\n} );
    } else {
        $fileserver = $resultset->create({
            cell => $cell->id,
            name => $args{name},
        });
        STDOUT->print( qq{Defined fileserver '$fileserver' in cell '$cell'\n} );
    }

    return $fileserver;

}

sub undef {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        cell => { type => SCALAR },
        name => { type => SCALAR },
    });

    my $resultset = $class->get_resultset;

    my $fileserver = $resultset->retrieve_bynames_assert(
        cell => $args{cell},
        name => $args{name},
    );

    if ( my @fileshares = $fileserver->fileshares ) {
        die(
            qq{Unable to undef fileserver '$fileserver', },
            qq{as it has the following fileshares:\n\t},
            join(qq{\n\t}, map { $_->name } @fileshares ), qq{\n},
        );
    }

    my $object_id = $fileserver->object_id;

    $fileserver->attrs->delete_all;
    $fileserver->delete;

    STDOUT->print( qq{Undefined fileserver '$object_id'\n} );

    return 1;

}

sub display {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        cell => { type => SCALAR, optional => 1 },
        name => { type => SCALAR, depends => [qw( cell )], optional => 1 },
    });

    my $resultset = $class->get_resultset;

    my @fileservers = ();

    if ( not $args{cell} ) {

        @fileservers = $resultset->all;

    } elsif ( not $args{name} ) {

        my $cell_rs = $class->get_resultset( q{Cell} );
        my $cell = $cell_rs->retrieve_bynames_assert( name => $args{cell} );
        @fileservers = $cell->fileservers;

    } else {

        $fileservers[0] = $resultset->retrieve_bynames_assert(
            cell => $args{cell},
            name => $args{name},
        );

    }

    $class->display_objects( objects => \@fileservers );

    if ( scalar @fileservers != 1 ) {
        return 1;
    }

    my $fileserver = $fileservers[0];

    if ( my @attrs = $fileserver->attrs ) {
        $attrs[0]->display_objects( objects => \@attrs );
    }

    if ( my @fileshares = $fileserver->fileshares ) {
        $fileshares[0]->display_objects(
            objects         => \@fileshares,
            display_columns => [qw( fileserver name )],
        );
    }

    return 1;

}

sub history {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        cell => { type => SCALAR },
        name => { type => SCALAR },
        %EFS::CLI::History::validation_specs,
    });

    $args{fileserver} = delete $args{name};

    return EFS::CLI::History->history(%args);

}

sub execlinks {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        cell  => { type => SCALAR },
        name  => { type => SCALAR },
        force => { type => SCALAR | UNDEF, optional => 1 },
    });

    $class->assert_maintenance_mode;

    my $resultset = $class->get_resultset;

    my $fileserver = $resultset->retrieve_bynames_assert(
        cell => $args{cell},
        name => $args{name},
    );

    $fileserver->cell->assert_fstype( q{nfs} );

    return EFS::RPC->execute(
        cell   => $fileserver->cell,
        class  => $class,
        method => q{execlinks_rpc},
        args   => [ %args ],
    );

}

sub execlinks_rpc {

    my $class = shift;
    my %args  = @_;

    my $resultset = $class->get_resultset;

    my $fileserver = $resultset->retrieve_bynames_assert(
        cell => $args{cell},
        name => $args{name},
    );

    my @fileqtrees = grep { $_->has_exectype } $fileserver->fileqtrees;

    my $fm_fileqtree = EFS::ForkManager->new({

        list => \@fileqtrees,

        action => sub {

            my $fileqtree = shift;

            return EFS::CLI::Fileqtree->execlinks_rpc(
                cell       => $fileqtree->cell->name,
                fileserver => $fileqtree->fileserver->name,
                fileshare  => $fileqtree->fileshare->name,
                name       => $fileqtree->name,
                force      => $args{force},
            );

        },

    });

    $fm_fileqtree->run ||
        die(
            qq{Unable to update execlinks on the following fileqtrees:\n\t},
            join( qq{\n\t}, map { $_->object_id } $fm_fileqtree->get_failed ),
            qq{\n},
        );

    STDOUT->print( qq{Updated execlinks on all fileqtrees\n} );

    return 1;

}

sub showmount {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        cell     => { type => SCALAR, optional => 1 },
        name     => { type => SCALAR, depends => [ q{cell} ], optional => 1 },
        verbose  => { type => BOOLEAN, optional => 1 },
        celltype => { type => SCALAR, optional => 1 },
    });

    my $resultset = $class->get_resultset;

    my $cell_rs = $class->get_resultset( q{Cell} );

    my @fileservers     = ();

    if ( $args{name} ) {

        $fileservers[0] = $resultset->retrieve_bynames_assert(
            cell => $args{cell},
            name => $args{name},
        );

        $fileservers[0]->cell->assert_fstype( q{nfs} );

    } elsif ( $args{cell} ) {

        my $cell = $cell_rs->retrieve_bynames_assert( name => $args{cell} );
        $cell->assert_fstype( q{nfs} );
        @fileservers = $cell->fileservers;

    } elsif ( $args{celltype} ) {

        my @cells = $cell_rs->search({
            celltype => $args{celltype},
            fstype   => q{nfs},
            hidden   => 0,
        });

        @fileservers = map { $_->fileservers } @cells;

    } else {

        my @cells = (
            $cell_rs->search({ celltype => q{prod}, fstype => q{nfs}, hidden => 0 }),
            $cell_rs->search({ celltype => q{dev},  fstype => q{nfs}, hidden => 0 }),
        );

        @fileservers = map { $_->fileservers } @cells;

    }

    if ( not @fileservers ) {
        die qq{No fileservers found that match the given criteria.\n};
    }

    my $results = {};
    my $stats   = {};
    my $summary = {};

    my $globalerrors = 0;

    foreach my $fileserver ( @fileservers ) {

        my $cell      = $fileserver->cell;
        my $cell_name = $cell->name;
        my $cell_type = $cell->celltype;

        my $fileserver_name = $fileserver->name;

        my @showmount = qx{ showmount -a $fileserver_name };

        if ( $CHILD_ERROR ) {
            STDOUT->print(
                qq{Unable to query fileserver '$fileserver' in },
                qq{cell '$cell', because\n},
                qq{the showmount command failed.\n}
            );
            $globalerrors++;
            next;
        }

        STDOUT->print(
            qq{Querying NFS clients of fileserver '$fileserver_name' in },
            qq{cell '$cell_name'\n}
        );

        my $fileserver_results =
            $results->{ $cell_name }->{ $fileserver_name } = {};
        my $fileserver_stats =
            $stats->{ $cell_name }->{ $fileserver_name } = {};
        my $fileserver_summary =
            $summary->{ $cell_name }->{ $fileserver_name } = {};

        foreach my $showmount ( @showmount ) {

            my ($clientname,$filesystem) = split( m/:/msx, $showmount, 2);

            # Skip the headers...  NOTE: Can't use --no-headers, as
            # that's not supported on Solaris (think cross-platform)
            next unless $filesystem =~ m{\A /vol/efs_}imsx;

            if ( $clientname =~ m/\A (\d+ \. \d+ \. \d+ \. \d+) \z/msx ) {
                $clientname     = hostname_of($clientname);
            }

            if ( $clientname =~ m{\A m \w+ arx \d+}msx ) {
                # Ignore the Acopia switches as clients
                next;
            }

            $fileserver_results->{ $clientname }->{ filesystems }->{ $filesystem }++;

        }

        foreach my $clientname ( sort keys %{ $fileserver_results } ) {

            my $client_results = $fileserver_results->{ $clientname };

            $client_results->{ hosttype } = $cell_type;
            $client_results->{ instance } = { unknown => 1 };

            my @filesystems = keys %{ $client_results->{ filesystems } };

            foreach my $filesystem ( @filesystems ) {

                if ( $filesystem =~ m{efs_qtree/ ([^/]+) /dist}imsx ) {
                    $client_results->{ instance }->{ $1 }++;
                    delete $client_results->{ instance }->{ unknown };
                }

            }

            if ( scalar keys %{ $client_results->{ instance } } > 1 ) {
                #
                # OK, this is really wierd...  We handle multiple
                # instances on a single client as follows.  For each
                # of the instance values, get a list of ALL of the
                # backwards and emulated instances, and delete them
                # from the hash.  The last one left will be the newest
                # instance name.  This will get the result wrong if
                # someone DOWNGRADES a box, but when we start to audit
                # the clients, we'll check this explictly, and this
                # nonsense which noone but Phil will ever grok can go
                # away.
                #
                foreach my $instance ( keys %{ $client_results->{instance} } ) {

                    my $instance_rs = $class->get_resultset(q{PlatformInstance});

                    my ($object) = $instance_rs->search({ name => $instance });

                    if ( $object ) {
                        foreach my $children ( $object->children_all ) {
                            delete $client_results->{ instance }->{ $children->name };
                        }
                    }

                }
            }

            if ( $client_results->{ instance }->{ unknown } ) {

                my $efsserver_rs = $class->get_resultset( q{EFSServer} );

                if ( $efsserver_rs->search({ name => $clientname }) ) {
                    # EFSServers mount other cell's filesystems, so if
                    # they appear to have an instance of unknown, then
                    # these are the /.efsadmin namespace mounts, to skip
                    # these.
                    delete $fileserver_results->{ $clientname };
                    next;
                }

            }

            if ( scalar keys %{ $client_results->{ instance } } > 1 ) {
                warn(
                    q{Unable to determine unique instance name },
                    qq{for $clientname in $cell_name\n\t},
                    join( qq{\n\t}, keys %{ $client_results->{ instance } } ),
                    qq{\n},
                );
            }

            $fileserver_stats->{ Hosttype }
                ->{ $client_results->{ hosttype } }++;

            foreach my $instance ( keys %{ $client_results->{ instance } } ) {
                $fileserver_stats->{ Instance }->{ $instance }++;
                 $fileserver_summary->{ $client_results->{ hosttype } }->{ $instance }++;
            }

        }

    }

    my $tables = {};

    $tables->{ verbose } = Text::FormatTable->new(q{l  r  r  r  l});
    $tables->{ verbose }->head(qw(Cell Fileserver Hostname Hosttype Instance));
    $tables->{ verbose }->rule(q{=});

    foreach my $cell_name ( sort keys %{ $results } ) {

        my $cell_results = $results->{ $cell_name };

        foreach my $fileserver_name ( sort keys %{ $cell_results } )  {

            my $fileserver_results = $cell_results->{ $fileserver_name };

            foreach my $clientname ( keys %{ $fileserver_results } ) {

                my $client_results = $fileserver_results->{ $clientname };
                my $instance = join( q{,}, keys %{ $client_results->{ instance } } );

                $tables->{ verbose }->row( $cell_name,
                                           $fileserver_name,
                                           $clientname,
                                           $client_results->{ hosttype },
                                           $instance || q{unknown});

            }

        }

    }

    if ( $args{verbose} ) {
        STDOUT->print($tables->{ verbose }->render($SCREEN_WIDTH));
    }

    $tables->{ summary } = Text::FormatTable->new(q{l  r  r  r  r});
    $tables->{ summary }->head( qw( Cell Fileserver HostType Instance Count ) );
    $tables->{ summary }->rule( q{=} );

    foreach my $cell_name ( sort keys %{ $summary } ) {

        my $cell_summary = $summary->{ $cell_name };

        foreach my $fileserver_name ( sort keys %{ $cell_summary } ) {

            my $fileserver_summary = $cell_summary->{ $fileserver_name };

            foreach my $hosttype ( sort keys %{ $fileserver_summary } ) {

                my $hosttype_summary = $fileserver_summary->{ $hosttype };

                foreach my $instance ( sort keys %{ $hosttype_summary } ) {

                    my $count = $hosttype_summary->{ $instance };

                    $tables->{ summary }->row( $cell_name,
                                               $fileserver_name,
                                               $hosttype,
                                               $instance,
                                               $count );

                }

            }

        }

    }

    STDOUT->print( qq{\n\n} );
    STDOUT->print( $tables->{ summary }->render($SCREEN_WIDTH) );

    foreach my $type ( qw( Hosttype Instance ) ) {
        $tables->{ $type } = Text::FormatTable->new(q{l  r  r  r});
        $tables->{ $type }->head( q{Cell}, q{Fileserver}, $type, q{Count} );
        $tables->{ $type }->rule( q{=} );
    }

    my $totals = {
        Hosttype => {},
        Instance => {},
        Cell     => {},
    };

    foreach my $cell_name ( sort keys %{ $stats } ) {

        my $cell_stats = $stats->{ $cell_name };

        foreach my $fileserver_name ( sort keys %{ $cell_stats } ) {

            my $fileserver_stats = $cell_stats->{ $fileserver_name };

            foreach my $type ( qw( Hosttype Instance ) ) {

                my @keys = reverse sort {
                    $fileserver_stats->{ $type }->{ $a } <=>
                        $fileserver_stats->{ $type }->{ $b }
                } keys %{ $fileserver_stats->{ $type } };

                foreach my $key ( @keys ) {

                    my $count = $fileserver_stats->{ $type }->{ $key };

                    $tables->{ $type }->row( $cell_name,
                                             $fileserver_name,
                                             $key,
                                             $count);

                    $totals->{ $type }->{ $key } += $count;

                    # Calculate the cell totals as a side effect of
                    # counting the hosttypes
                    if ( $type eq q{Hosttype} ) {
                        $totals->{ Cell }->{ $cell_name } += $count;
                    }

                }

            }

        }

    }

    foreach my $type ( qw( Hosttype Instance ) ) {
        STDOUT->print(qq{\n\n});
        STDOUT->print( $tables->{ $type }->render($SCREEN_WIDTH) );
    }

    if ( scalar keys %{ $stats } > 1 ) {

        STDOUT->print(qq{\n\n});
        STDOUT->print(qq{Global Totals:\n});

        foreach my $type ( qw( Hosttype Instance Cell ) ) {

            my $table = Text::FormatTable->new(q{r  r});
            my $total = 0;

            $table->head( $type, q{Count} );
            $table->rule(q{=});

            my @totals = reverse sort {
                $totals->{ $type }->{ $a } <=> $totals->{ $type }->{ $b };
            } keys %{ $totals->{ $type } };

            foreach my $key ( sort keys %{ $totals->{ $type } } ) {
                my $count = $totals->{ $type }->{ $key };
                $table->row( $key, $count );
                $total += $count;
            }

            $table->rule( q{-} );
            $table->row( q{TOTAL}, $total );

            STDOUT->print( $table->render($SCREEN_WIDTH) );
            STDOUT->print( qq{\n\n} );

        }

    }

    if ( $globalerrors ) {
        die qq{Unable to query all data sources.  Results may be incomplete.\n};
    }

    return 1;

}

{

    my %hostname_of = ();

    sub hostname_of {

        my $address = shift;

        if ( $hostname_of{ $address } ) {
            return $hostname_of{ $address };
        }

        my $backwards = join( q{.}, reverse split( m/\./msx, $address ) );

        my @nslookup = qx{ nslookup $address };

        my $name = $address;

        foreach my $line ( @nslookup ) {

            if ( $line =~ m/\A $backwards \. in-addr \. arpa \h+ name \h+ = \h+ (\S+)/msx ) {
                $name = $1;
                last;
            } elsif ( $line =~ m/\A Name: \h+ (\S+)/msx ) {
                $name = $1;
                last;
            }

        }

        $name =~ s/\. \z//msx;
        $name = lc($name);

        # This appears to be how unused addresses are marked in DNS...
        if ($name eq q{open}) {
            $name = $address;
        }

        return $hostname_of{ $address } = $name;

    }

}

1;

# Local Variables:
#   mode: cperl
#   cperl-indent-level: 4
#   fill-column: 78
# End:
# vim: expandtab shiftwidth=4:
