package EFS::CLI::GroupRights;

use strict;
use warnings;
use English;

use base q{EFS::CLI};
use base q{EFS::DBI};
use Params::Validate qw( :all );

sub define {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        metaproj  => { type => SCALAR },
        role      => { type => SCALAR },
        groupname => { type => SCALAR },
    });

    my $metaproj_rs = $class->get_resultset( q{Metaproj} );

    my $metaproj = $metaproj_rs->retrieve_bynames_assert(
        name => $args{metaproj}
    );

    if ( not $args{role} ~~ [qw( dev qa prod auto admin )] ) {
        die(
            qq{Invalid role '$args{role}'\n},
            qq{Must be one of: dev, qa, prod, auto or admin\n},
        );
    }

    my $groupname = EFS::Groupname->new({ name => $args{groupname} });

    my $resultset = $class->get_resultset;

    my ($grouprights) = $resultset->search({
        metaproj  => $metaproj->id,
        role      => $args{role},
        groupname => $groupname,
    });

    if ( $grouprights ) {

        STDOUT->print(
            qq{Groupname $groupname already has },
            qq{$args{role} rights to $metaproj\n},
        );

    } else {

        $grouprights = $resultset->create({
            metaproj  => $metaproj->id,
            role      => $args{role},
            groupname => $groupname,
        });

        STDOUT->print(
            qq{Granted $groupname $args{role} rights to $metaproj\n},
        );

    }

    return 1;

}

sub undef {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        metaproj  => { type => SCALAR },
        role      => { type => SCALAR },
        groupname => { type => SCALAR },
    });

    my $resultset = $class->get_resultset;

    my $grouprights = $resultset->retrieve_bynames_assert(
        metaproj  => $args{metaproj},
        role      => $args{role},
        groupname => $args{groupname},
    );

    $grouprights->delete;

    STDOUT->print(
        qq{Removed $args{groupname} $args{role} rights to $args{metaproj}\n},
    );

    return 1;

}

sub display {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        metaproj  => { type => SCALAR, optional => 1 },
        role      => { type => SCALAR, depends  => [qw( metaproj )], optional => 1 },
        groupname => { type => SCALAR, depends  => [qw( role )], optional => 1 },
    });

    my $metaproj_rs = $class->get_resultset( q{Metaproj} );

    my $metaproj = $metaproj_rs->retrieve_bynames_assert(
        name => $args{metaproj},
    );

    my $resultset = $class->get_resultset;

    my @grouprights = ();

    if ( not $args{metaproj} ) {

        @grouprights = $resultset->all;

    } elsif ( not $args{role} ) {

        @grouprights = $resultset->search({ metaproj => $metaproj->id });

    } elsif ( not $args{groupname} ) {

        @grouprights = $resultset->search({
            metaproj => $metaproj->id,
            role     => $args{role},
        });

    } else {

        $grouprights[0] = $resultset->retrieve_bynames_assert(
            metaproj  => $args{metaproj},
            role      => $args{role},
            groupname => $args{groupname},
        );

    }

    if ( @grouprights ) {
        $class->display_objects( objects => \@grouprights );
    } else {
        STDOUT->print(
            qq{No grouprights match the specified criteria.\n},
        );
    }

    return 1;

}

1;

# Local Variables:
#   mode: cperl
#   cperl-indent-level: 4
#   fill-column: 78
# End:
# vim: expandtab shiftwidth=4:
