package EFS::CLI::Location;

use strict;
use warnings;
use English;

use base q{EFS::CLI};
use base q{EFS::DBI};
use Params::Validate qw( :all );

sub define {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        region      => { type => SCALAR },
        campus      => { type => SCALAR },
        name        => { type => SCALAR },
        description => { type => SCALAR },
    });

    my $campus_rs = $class->get_resultset( q{Campus} );

    my $campus = $campus_rs->retrieve_bynames_assert(
        region => $args{region},
        name   => $args{campus},
    );

    my $resultset = $class->get_resultset;

    my ($location) = $resultset->search({ name => $args{name} });

    if ( $location ) {

        my $campus_name = $location->campus->name;
        my $region_name = $location->region->name;

        if ( $campus_name ne $args{campus} ) {
            die(
                qq{Location $args{name} is already defined, but in },
                qq{campus $campus_name, region $region_name\n},
            );
        } else {
            STDOUT->print(
                qq{Location $region_name/$campus_name/$args{name} },
                qq{is already defined\n},
            );
        }

        if ( $location->description ne $args{description} ) {
            $location->set_column( description => $args{description} );
            $location->update;
            STDOUT->print(
                qq{Updated description of location $location\n},
            );
        }

    } else {

        $location = $resultset->create({
            campus      => $campus->id,
            name        => $args{name},
            description => $args{description},
        });

        STDOUT->print(
            qq{Defined location $args{region}/$args{campus}/$args{name}\n},
        );

    }

    return $location;

}

sub undef {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        region => { type => SCALAR },
        campus => { type => SCALAR },
        name   => { type => SCALAR },
    });

    my $resultset = $class->get_resultset;

    my $location = $resultset->retrieve_bynames_assert(
        region => $args{region},
        campus => $args{campus},
        name   => $args{name},
    );

    if ( my @cells = $location->cells ) {
        die(
            qq{Unable to undef location $location, because it has\n},
            qq{the following cells:\n\t},
            join(qq{\n\t}, map { $_->name } @cells ), qq{\n},
        );
    }

    if ( my @pods = $location->pods ) {
        die(
            qq{Unable to undef location $location, because it has\n},
            qq{the following pods:\n\t},
            join(qq{\n\t}, map { $_->name } @pods ), qq{\n},
        );
    }

    if ( my @cabinets = $location->cabinets ) {
        die(
            qq{Unable to undef location $location, because it has\n},
            qq{the following cabinets:\n\t},
            join(qq{\n\t}, map { $_->name } @cabinets ), qq{\n},
        );
    }

    my $object_id = $location->object_id;

    $location->attrs->delete_all;
    $location->delete;

    STDOUT->print( qq{Undefined location $object_id\n} );

    return 1;

}

sub display {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        region => { type     => SCALAR,
                    optional => 1 },
        campus => { type     => SCALAR,
                    depends  => [qw( region )],
                    optional => 1 },
        name   => { type     => SCALAR,
                    depends  => [qw( campus)],
                    optional => 1 },
    });

    my $resultset = $class->get_resultset;

    my @locations = ();

    if ( not $args{region} ) {

        @locations = $resultset->all;

    } elsif ( not $args{campus} ) {

        my $region_rs = $class->get_resultset( q{Region} );

        my $region = $region_rs->retrieve_bynames_assert(
            name => $args{region},
        );

        @locations = $region->locations;

    } elsif ( not $args{name} ) {

        my $campus_rs = $class->get_resultset( q{Campus} );

        my $campus = $campus_rs->retrieve_bynames_assert(
            region => $args{region},
            name   => $args{campus},
        );

        @locations = $campus->locations;

    } else {

        $locations[0] = $resultset->retrieve_bynames_assert(
            region => $args{region},
            campus => $args{campus},
            name   => $args{name},
        );

    }

    if ( not @locations ) {
        STDOUT->print( qq{No locations match the specified criteria\n} );
        return 1;
    }

    $class->display_objects(
        objects => [ $class->sort_objects( @locations ) ],
    );

    if ( scalar @locations != 1 ) {
        return 1;
    }

    my $location = $locations[0];

    if ( my @attrs = $location->attrs ) {
        $attrs[0]->display_objects( objects => \@attrs );
    }

    if ( my @cells = $location->cells ) {
        $cells[0]->display_objects( objects => \@cells );
    }

    if ( my @pods = $location->pods ) {
        $pods[0]->display_objects( objects => \@pods );
    }

    if ( my @cabinets   = $location->cabinets ) {
        $cabinets[0]->display_objects( objects => \@cabinets );
    }

    return 1;

}

sub history {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        region => { type => SCALAR },
        campus => { type => SCALAR },
        name   => { type => SCALAR },
        %EFS::CLI::History::validation_specs,
    });

    $args{location} = delete $args{name};

    return EFS::CLI::History->history(%args);

}

1;

# Local Variables:
#   mode: cperl
#   cperl-indent-level: 4
#   fill-column: 78
# End:
# vim: expandtab shiftwidth=4:
