package EFS::DBI::Result::Cell;

use strict;
use warnings;
use English;

use base q{EFS::DBI::Result};

use EFS::FQDN qw( fqdn );

use Readonly;
use List::Util qw( shuffle );
use List::MoreUtils qw( uniq );
use Params::Validate qw( :all );
use File::Basename qw( basename dirname );

__PACKAGE__->load_components( qw(+EFS::DBI::Prefix Core) );

__PACKAGE__->table( q{cell} );

__PACKAGE__->add_columns( qw( name location celltype fstype
                              parentcell disabled hidden ) );
__PACKAGE__->add_columns(
    id => { __PACKAGE__->get_id_column_attrs },
);
__PACKAGE__->set_primary_key( q{id} );
__PACKAGE__->sequence( q{cell_seq} );

__PACKAGE__->belongs_to( location    => q{EFS::DBI::Result::Location} );
__PACKAGE__->belongs_to( parentcell  => q{EFS::DBI::Result::Cell} );

__PACKAGE__->has_many( fileservers => q{EFS::DBI::Result::Fileserver} );
__PACKAGE__->has_many( projects    => q{EFS::DBI::Result::Project} );
__PACKAGE__->has_many( homes       => q{EFS::DBI::Result::Home} );

__PACKAGE__->has_many( efsserver_cell => q{EFS::DBI::Result::EFSServerCell}, q{cell} );
__PACKAGE__->many_to_many( q{efsservers} => q{efsserver_cell}, q{efsserver} );

__PACKAGE__->has_many( attrs => q{EFS::DBI::Result::CellAttrs} );

Readonly my $SCREEN_WIDTH => 1024;

sub campus {
    return shift->location->campus;
}

sub region {
    return shift->campus->region;
}

sub fileshares {
    return map { $_->fileshares(@_) } shift->fileservers;
}

sub fileqtrees {
    return map { $_->fileqtrees(@_) } shift->fileshares;
}

sub efsservers_byhosttype {
    my $self = shift;
    my $hosttype = shift;
    return grep { $_->hosttype eq $hosttype } $self->efsservers;
}

sub children {
    my $self = shift;
    my $resultset = $self->get_resultset( q{Cell} );
    my @children = $resultset->search({ parentcell => $self->id });
    return @children;
}

sub devcell {

    my $self = shift;

    if ( $self->celltype eq q{dev} ) {
        return $self;
    } elsif ( $self->parentcell->celltype eq q{dev} ) {
        # $self is a prod or test cell with a dev parentcell, so
        # return the parent
        return $self->parentcell;
    } else {
        # $self is a prod cell with a prod parentcell, so
        # return the grandparent
        return $self->parentcell->parentcell;
    }

}

sub object_id {
    return shift->name;
}

Readonly my $REGION_MAXLENGH   => 24;
Readonly my $CAMPUS_MAXLENGH   => 64;
Readonly my $CELLTYPE_MAXLENTH => 8;
Readonly my $LOCATION_MAXLENTH => 128;

sub sort_id {
    my $self = shift;
    return join( qq{\n},
                 sprintf( q{%-*s}, $REGION_MAXLENGH,
                          $self->region->name ),
                 sprintf( q{%-*s}, $CAMPUS_MAXLENGH,
                          $self->campus->name ),
                 sprintf( q{%-*s}, $CELLTYPE_MAXLENTH,
                          $self->celltype ),
                 sprintf( q{%-*s}, $LOCATION_MAXLENTH,
                          $self->location->name ) ) . qq{\n};
}

# Wrapper functions for define/undef/display, to allow for a simpler
# specification of the cell

sub shortname {
    my $self = shift;
    my $name = ref $self ? $self->name : shift;
    return join( q{.}, ( split( m/\./msx, $name ) )[0,1] );
}

sub domain {
    my $self = shift;
    ( my $domain = $self->name ) =~ s/\A [^.]+ \.//gmsx;
    return $domain;
}

sub assert_cell_is_dev {
    my $self = shift;
    if ( $self->celltype ne q{dev} ) {
        die qq{Cell $self is not a dev cell\n};
    }
    return 1;
}

sub assert_supports_filetype {
    my $self = shift;
    my $filetype = shift;
    $self->supports_filetype( $filetype ) ||
        die qq{Cell $self doesn't support filetype $filetype\n};
    return 1;
}

sub supports_filetype {
    my $self = shift;
    my $filetype = shift;
    my ($fileshare) = $self->fileshares({ filetype => $filetype });
    return ( $fileshare ? 1 : 0 );
}

sub assert_fstype {
    my $self = shift;
    my $fstype = shift;
    if ( $fstype ne $self->fstype ) {
        my $name = $self->name;
        die qq{Cell $name does not have fstype $fstype\n};
    }
    return 1;
}

# utility subroutines to compare two cells by celltype.
#   1. dev < prod < test
sub _celltype_importance {
    ##  no critic (MagicNumbers)
    my $self = shift;
    my %types_ordinality = (
        dev  => 1,
        prod => 2,
        test => 3,
    );
    if (my $ordinality = $types_ordinality{ $self->celltype } ) {
        return $ordinality;
    }
    else {
        # Odd... shouldn't get here
        warn( qq{Cell $self has weird type }, $self->celltype, qq{...\n} );
        return 4;
    }
}

#   2. prod > test
sub _parenttype_importance {
    ##  no critic (MagicNumbers)
    my $self = shift;
    if ( not defined $self->parentcell ) {
        return 1;
    }
    if ( $self->celltype eq q{dev} ) {
        # Shouldn't happen; included for completeness
        die qq{Cell $self is dev but has parent?\n};
    }
    if ( $self->celltype eq q{prod} ) {
        return 2;
    }
    if ( $self->celltype eq q{test} ) {
        return 3;
    }
    # Odd... shouldn't get here
    die qq{Cell $self seems to have a parent but is of no known type...\n};
    return;
}

sub find_efsserver {

    my $self = shift;

    my $fqdn = fqdn;

    my $efsserver_rs = $self->get_resultset( q{EFSServer} );

    my $localhost = $efsserver_rs->retrieve_bynames_assert( name => $fqdn );

    if ( $localhost->supports_cell( $self ) ) {
        return $localhost;
    }

    my ($efsserver) = shuffle $self->efsservers_byhosttype( $localhost->hosttype );

    if ( not $efsserver ) {
        die qq{Unable to locate an EFSServer for cell $self\n};
    }

    return $efsserver;

}

sub admin_path {
    my $self = shift;
    if ( $self->fstype eq q{nfs} ) {
        return $self->admin_path_nfs( @_ );
    } else {
        return $self->admin_path_afs( @_ );
    }
}

sub admin_path_nfs {

    my $self = shift;

    my (%args) = validate( @_, {
        type => { type => SCALAR },
    });

    my ($fileshare) = $self->fileshares({ filetype => $args{type} }) or
        die(qq{Unable to determine $args{type} fileshare for $self});

    my ($fileqtree) = $fileshare->fileqtrees;

    my $pathname = join( q{/},
                         q{/.efsadmin},
                         $self->name,
                         $fileshare->fileserver->name,
                         $fileshare->name,
                         $fileqtree->name );

    $pathname =~ s{/+}{/}gmsx;

    return $pathname;

}

sub admin_path_afs {

    my $self = shift;

    my (%args) = validate( @_, {
        type => { type => SCALAR },
    });

    my $name = $self->name;

    return qq{/afs/$name/$args{type}};

    # XXX: That will work for dev and home, but what about dist?  That's much
    # harder, since dev and home aren't replicated, but dist is.

}

#
# XXX: This is a total hack -- wait for the flexcache timeout to
# expire.  This has to get yanked for the OSS version, or otherwise
# rethought.
#
sub wait_for_dist_cache {
    my $self = shift;
    return if $self->fstype eq q{afs};
    ##  no critic (MagicNumbers)
    return sleep(10);
}

sub _generate_cell_auto_map_global {

    my $self        = shift;
    my $install_dir = shift;

    $self->assert_fstype( q{nfs} );

    my $cell_name = $self->name;

    EFS::Util->MakeDir( path => $install_dir );
    EFS::Util->MakeDir( path => qq{$install_dir/$cell_name} );

    my $fileshare_rs = $self->get_resultset( q{Fileshare} );

    my @fileshares = $fileshare_rs->all;

    my %table = ();

    foreach my $fileshare ( @fileshares ) {
        if ( $fileshare->filetype ne q{dist} ) {
            $table{ $fileshare->autoname } =
                join( q{:}, $fileshare->fileserver->name, $fileshare->name );
        }
    }

    my @nfs_options = qw( rw intr tcp vers=3 );

    my $parameter_rs   = $self->get_resultset( q{Parameter} );
    my $nfs_options_rw = $parameter_rs->configured( q{nfs_options_rw} );
    push @nfs_options, $nfs_options_rw if $nfs_options_rw;

    my $nfs_options = join q{,}, @nfs_options;

    my $table = Text::FormatTable->new(q{l  l  r});

    foreach my $key ( sort keys %table ) {
        $table->row( $key, qq{-$nfs_options}, $table{$key} );
    }

    my $content = $table->render($SCREEN_WIDTH);

    my $file = qq{$install_dir/$cell_name/global.map};

    EFS::Config->update_file(
        file    => $file,
        content => $content,
        message => qq{Generated autofs map file $file\n},
    );

    return 1;

}

sub _generate_cell_auto_map_admin {

    my $self        = shift;
    my $install_dir = shift;

    $self->assert_fstype( q{nfs} );

    my $cell_name = $self->name;

    my $dist;

    if ( $install_dir =~ m{\A /var}msx ) {
        $dist = qq{$install_dir/$cell_name};
    } else {
        $dist = qq{/efs/dist/efs/config-automaps/current/common/autofs/$cell_name};
    }

    EFS::Util->MakeDir( path => $install_dir );
    EFS::Util->MakeDir( path => qq{$install_dir/$cell_name} );

    my $admin = qq{$install_dir/$cell_name/admin};

    EFS::Util->MakeDir( path => $admin );

    # This is the list of cells that share efsservers with
    # this cell.  We have to include ALL of the fileshares for
    # these others cells, too.
    my @cells      = uniq map { $_->cells } $self->efsservers;
    my @fileshares = map { $_->fileshares } @cells;

    my %data = ();

    foreach my $volume_share ( @fileshares ) {
        $data{ $volume_share->cell->name }->
            { $volume_share->fileserver->name }->
                { $volume_share->name } = $volume_share;
    }

    my $adminmap = Text::FormatTable->new(q{l  l  r});

    foreach my $othercell ( sort keys %data ) {

        $adminmap->row( $othercell, q{-fstype=autofs},
                        qq{file:$dist/admin/$othercell.map} );

        EFS::Util->MakeDir( path => qq{$admin/$othercell} );

        my $cellmap = Text::FormatTable->new(q{l  l  r});

        my @fileservers = sort keys %{ $data{$othercell} };

        foreach my $fileserver ( @fileservers ) {

            $cellmap->row( $fileserver, q{-fstype=autofs},
                           qq{file:$dist/admin/$othercell/$fileserver.map} );

            EFS::Util->MakeDir( path => qq{$admin/$othercell/$fileserver} );

            my $volmap = Text::FormatTable->new(q{l  l  r});

            my @volumes = sort keys %{ $data{$othercell}->{$fileserver} };

            foreach my $volume ( @volumes ) {

                my $volume_share = $data{$othercell}->{$fileserver}->{$volume};
                my $volname      = basename($volume);

                $volmap->row( $volname, q{-rw,intr,tcp,vers=3},
                              join( q{:},
                                    $volume_share->fileserver->name,
                                    $volume_share->name ) );

            }

            EFS::Config->update_file(
                file    => qq{$admin/$othercell/$fileserver/vol.map},
                content => $volmap->render($SCREEN_WIDTH),
            );

            my $fsmap = Text::FormatTable->new(q{l  l  r});
            $fsmap->row( q{vol}, q{-fstype=autofs},
                         qq{file:$dist/admin/$othercell/$fileserver/vol.map} );

            EFS::Config->update_file(
                file    => qq{$admin/$othercell/$fileserver.map},
                content => $fsmap->render($SCREEN_WIDTH),
            );

        }

        EFS::Config->update_file(
            file    => qq{$admin/$othercell.map},
            content => $cellmap->render($SCREEN_WIDTH),
        );

    }

    EFS::Config->update_file(
        file    => qq{$install_dir/$cell_name/admin.map},
        content => $adminmap->render($SCREEN_WIDTH),
    );

    return 1;

}

sub _generate_cell_amd_map_global {

    my $self        = shift;
    my $install_dir = shift;

    $self->assert_fstype( q{nfs} );

    my $cell_name = $self->name;

    EFS::Util->MakeDir( path => $install_dir );
    EFS::Util->MakeDir( path => qq{$install_dir/$cell_name} );

    my $fileshare_rs = $self->get_resultset( q{Fileshare} );

    my @fileshares = $fileshare_rs->all;

    my %table = ();

    foreach my $fileshare ( @fileshares ) {
        if ( $fileshare->filetype ne q{dist} ) {
            my $rhost = $fileshare->fileserver->name;
            my $rfs   = $fileshare->name;
            $table{ $fileshare->autoname } =
                join( q{;}, qq{rhost:=$rhost}, qq{rfs:=$rfs} );
        }
    }

    my $table = Text::FormatTable->new(q{l  r});

    $table->row( q{/defaults}, q{type:=nfs;opts:=rw,intr,tcp,nfs3} );

    foreach my $key ( sort keys %table ) {
        $table->row( $key, $table{$key} );
    }

    my $content = $table->render($SCREEN_WIDTH);

    my $file = qq{$install_dir/$cell_name/global.map};

    EFS::Config->update_file(
        file    => $file,
        content => $content,
        message => qq{Generated amd map file $file\n},
    );

    return 1;

}

sub _generate_cell_amd_map_admin {

    my $self        = shift;
    my $install_dir = shift;

    $self->assert_fstype( q{nfs} );

    my $cell_name = $self->name;

    EFS::Util->MakeDir( path => $install_dir );
    EFS::Util->MakeDir( path => qq{$install_dir/$cell_name} );

    # This is the list of cells that share efsservers with
    # this cell.  We have to include ALL of the fileshares for
    # these others cells, too.
    my @cells      = uniq map { $_->cells } $self->efsservers;
    my @fileshares = map { $_->fileshares } @cells;

    ## no critic: RequireInterpolationOfMetachars
    my $autovalue = join(
        q{;},
        q{type:=auto},
        q{opts:=browsable},
        q{fs:=${map}},
        q{pref:=${key}/},
    );

    my %table = ();

    my $table = Text::FormatTable->new(q{l  r});

    $table->row( q{/defaults}, q{type:=nfs;opts:=rw,intr,tcp,nfs3} );

    foreach my $fileshare ( @fileshares ) {

        my $fileshare_name  = $fileshare->name;
        my $cell_name       = $fileshare->cell->name;
        my $fileserver      = $fileshare->fileserver;
        my $fileserver_name = $fileserver->name;

        $table{ $cell_name }                      = $autovalue;
        $table{ qq{$cell_name/$fileserver_name} } = $autovalue;

        my $fileshare_dir = dirname $fileshare_name;

        while ( $fileshare_dir ne q{/} ) {
            $table{ qq{$cell_name/$fileserver_name$fileshare_dir} } = $autovalue;
            $fileshare_dir = dirname $fileshare_dir;
        }

        $table{ qq{$cell_name/$fileserver_name$fileshare_name} } =
            qq{rhost:=$fileserver_name;rfs:=$fileshare_name};

    }

    foreach my $key ( sort keys %table ) {
        $table->row( $key, $table{$key} );
    }

    EFS::Config->update_file(
        file    => qq{$install_dir/$cell_name/admin.map},
        content => $table->render($SCREEN_WIDTH),
    );

    return 1;

}

sub get_partition_list {

    my $self = shift;

    my $vos = AFS::Command::VOS->new;

    my @results = ();

    foreach my $fileserver ( $self->fileservers ) {
        my $result = $vos->partinfo(
            server => $fileserver->name,
            cell   => $fileserver->cell->name,
        );
        my ($partition) = sort { $b->available <=> $a->available } $result->getPartitions;
        push @results, [ $fileserver->name, $partition->partition ];
    }

    return shuffle @results;

}

1;

# Local Variables:
#   mode: cperl
#   cperl-indent-level: 4
#   fill-column: 78
# End:
# vim: expandtab shiftwidth=4:
