package EFS::DBI::Result::DistStats;

use strict;
use warnings;
use English;

use base q{EFS::DBI::Result};

use Params::Validate qw( :all );

__PACKAGE__->load_components( qw/+EFS::DBI::Prefix TimeStamp Core/ );

__PACKAGE__->table( q{diststats} );

__PACKAGE__->add_columns( qw( metaproj project release status
                              src_cell dest_cell file_count file_xfer_count
                              file_size file_xfer_size bytes_sent
                              bytes_received bytes_rate total_size speedup
                              history ) );
__PACKAGE__->add_columns(
    id => { __PACKAGE__->get_id_column_attrs },
);
__PACKAGE__->set_primary_key( q{id} );
__PACKAGE__->sequence( q{diststats_seq} );

__PACKAGE__->belongs_to( history => q{EFS::DBI::Result::History} );

__PACKAGE__->has_many( rsyncstats => q{EFS::DBI::Result::RsyncStats} );

sub rollup_rsyncstats {

    my $self = shift;

    my $elapsed_time = shift;         # Yeah, a bit of a hack...  Look
                                      # where this is called
    $elapsed_time ||= 1;              # Nothing takes less than a
                                      # second...

    my @rsyncstats = $self->rsyncstats;

    my $bytes_sent      = 0;
    my $bytes_received  = 0;
    my $total_size      = 0;
    my $file_xfer_count = 0;
    my $file_xfer_size  = 0;
    my $file_size       = 0;
    my $file_count      = 0;

    foreach my $rsyncstats ( @rsyncstats ) {

        $bytes_sent      += $rsyncstats->bytes_sent;
        $bytes_received  += $rsyncstats->bytes_received;
        $total_size      += $rsyncstats->total_size;
        $file_xfer_count += $rsyncstats->file_xfer_count;
        $file_xfer_size  += $rsyncstats->file_xfer_size;
        $file_size       += $rsyncstats->file_size;
        $file_count      += $rsyncstats->file_count;

    }

    $self->set_columns({
        bytes_sent      => $bytes_sent,
        bytes_received  => $bytes_received,
        total_size      => $total_size,
        file_xfer_count => $file_xfer_count,
        file_xfer_size  => $file_xfer_size,
        file_size       => $file_size,
        file_count      => $file_count,
        bytes_rate      => ( ($bytes_sent+$bytes_received) / $elapsed_time ),
    });

    if ( $bytes_sent + $bytes_received ) {
        $self->set_column(
            speedup => sprintf( q{%.2f}, ($total_size / ($bytes_sent + $bytes_received)) )
        );
    } else {
        $self->set_column( speedup => 0 );
    }

    return;
}

sub display_stats {

    my $self = shift;

    my (%args) = validate( @_, {
        query => { type => BOOLEAN, optional => 1 },
    });

    my $summary = qq{Distribution completed:\n};

    my $transfered = ( $args{query} ? q{Need to transfer} : q{Transfered} );

    $summary .= sprintf(
        qq{$transfered %d of %d total files, %d of %d total bytes\n},
        $self->file_xfer_count, $self->file_count,
        $self->file_xfer_size, $self->file_size
    );

    return STDOUT->print($summary);

}

1;

# Local Variables:
#   mode: cperl
#   cperl-indent-level: 4
#   fill-column: 78
# End:
# vim: expandtab shiftwidth=4:
