package EFS::DBI::Result::Release;

use strict;
use warnings;
use English;

use base q{EFS::DBI::Result};
use base q{EFS::DistableObject};

use Params::Validate qw( :all );

__PACKAGE__->load_components( qw/+EFS::DBI::Prefix Core/ );

__PACKAGE__->table( q{release} );

__PACKAGE__->add_columns(
    id            => { __PACKAGE__->get_id_column_attrs },
    project       => { efs_column_group => [qw( display )] },
    name          => { efs_column_group => [qw( display )] },
    stage         => { efs_column_group => [qw( display )] },
    locked        => { efs_column_group => [qw( display )] },
    destroyed     => { efs_column_group => [qw( display )] },
    devvolume     => { efs_column_group => [qw( optional )] },
    distvolume    => { efs_column_group => [qw( optional )] },
    noworldaccess => { efs_column_group => [qw( access )] },
    distgroup     => { efs_column_group => [qw( access )] },
    packaged      => {},
    uploaded      => {},
    downloaded    => {},
    cloneof       => {},
);

__PACKAGE__->set_primary_key( q{id} );
__PACKAGE__->sequence( q{release_seq} );

__PACKAGE__->belongs_to( devvolume  => q{EFS::DBI::Result::AFSVolume} );
__PACKAGE__->belongs_to( distvolume => q{EFS::DBI::Result::AFSVolume} );

__PACKAGE__->belongs_to( project => q{EFS::DBI::Result::Project} );
__PACKAGE__->belongs_to( cloneof => __PACKAGE__ );

__PACKAGE__->has_many( installs => q{EFS::DBI::Result::Install} );

__PACKAGE__->has_many( attrs => q{EFS::DBI::Result::ReleaseAttrs} );

sub clones {
    my $self = shift;
    my $resultset = $self->get_resultset;
    my @clones = $resultset->search({ cloneof => $self->id });
    return @clones;
}

sub unclone {
    my $self = shift;
    foreach my $install ( $self->installs ) {
        $install->unclone;
    }
    $self->set_column( cloneof => undef );
    $self->update;
    return 1;
}

sub display_columns {
    return qw(metaproj project name stage locked
              spaceused filecount dircount linkcount);
}

sub metaproj {
    use Carp;
    my $self = shift;
    my $project = $self->project || confess;
    return $project->metaproj;
}

sub _metaproj {
    return shift->project->metaproj;
}

sub cell {
    return shift->project->cell;
}

# Used for conformance to EFS::DistableObject superclass
# Release of a release is the release itself
sub release {
    return shift;
}

sub releasealiases {
    my $self = shift;
    my $resultset = $self->get_resultset(q{ReleaseAlias});
    my @releasealiases = $resultset->search({
        project => $self->project->id,
        target  => $self->name,
        @_,
    });
    return @releasealiases;
}

## Next two methods are for DistableObject compliance
sub subitems {
    return shift->installs;
}

sub container {
    return shift->project;
}

sub dependencies {

    my $self = shift;

    my %search_args = ( dependent_release => $self->id, @_ );

    my @dependencies = ();

    push ( @dependencies,
        $self->get_resultset(q{DependencyRelease})->search({ %search_args })
    );

    push ( @dependencies,
        $self->get_resultset(q{DependencyReleaseAlias})->search({ %search_args })
    );

    return @dependencies;

}

sub dependents {
    my $self = shift;
    my $resultset = $self->get_resultset(q{DependencyRelease});
    my @dependents = $resultset->search( { release => $self->id, @_ } );
    return @dependents;
}

sub releasealias_dependents {

    my $self = shift;

    my @releasealiases = $self->all_releasealiases;

    my @dependents = ();

    my $resultset = $self->get_resultset(q{DependencyReleaseAlias});

    foreach my $releasealias (@releasealiases) {
        push @dependents,
            $resultset->search( { releasealias => $releasealias->id, @_, } );
    }

    return @dependents;

}

sub all_dependents {
    my $self= shift;
    return ( $self->dependents(@_), $self->releasealias_dependents(@_), );

}

sub all_releasealiases {

    my $self = shift;

    my @results = ();

    my $resultset = $self->get_resultset(q{ReleaseAlias});

    my @releasealiases = $resultset->search( { project => $self->project->id, } );

    foreach my $releasealias (@releasealiases) {

        if ( not $releasealias->resolves_to_name( $self->name ) ) {
            next;
        }

        push @results, $releasealias;

    }

    return @results;

}

sub object_id {
    my $self = shift;
    my $separator = shift || q{/};
    return join( $separator,
                 $self->metaproj->name,
                 $self->project->name,
                 $self->name );
}

sub display_translate {

    my $class = shift;
    my ($key,$value) = @_;

    if ( $key eq q{locked} ) {
        return ( $value ? q{yes} : q{no} );
    }
    elsif ( $key eq q{spaceused} ) {
        return $class->size_shorthand($value);
    }
    else {
        return $value;
    }

}

sub get_diststate {
    return EFS::Diststate::Release->new( { release => shift } );
}

sub assert_not_disabled {

    my $self = shift;

    if ( my $disabled = $self->is_disabled ) {
        die qq{Release '$self' is disabled: $disabled\n};
    }

    return;
}

sub assert_not_destroyed {

    my $self = shift;

    if ( $self->destroyed ) {
        die(
            qq{The release '$self' has been destroyed.\n},
            q{It must be recovered before any },
            qq{further operations can be performed.\n}
        );
    }

    return;
}

sub get_path {
    my $self = shift;
    my $pathtype = shift || q{dist};
    return $self->project->get_path($pathtype) . q{/} . $self->name;
}

sub _run_lock_unlock_trigger {
    my $self = shift;
    my %args = @_;

    # No triggers except on unlock for now
    if ( not defined $args{lockmode} or $args{lockmode} ne q{unlock} ) {
        return 1;
    }

    # Only trigger on prod releases
    if ( not $self->has_prod_stages ) {
        return 1;
    }

    my @dependent_targets = map { $_->dependent_release } $self->all_dependents;

    return EFS::Notification->notify_mprs_template(
        %args,
        template      => q{notify_prod_unlock},
        template_args => {
            release    => $self->object_id,
            user       => $args{username},
            dependents => \@dependent_targets,
        },
        reply_to   => $self,
        recipients => [ $self, @dependent_targets ],
        priority   => 1,
    );

}

sub validate_audits {

    my $self = shift;

    my @installs = $self->installs;

    if ( not @installs ) {
        die qq{The release '$self' has no installs at all.\n};
    }

    STDOUT->print( qq{Validating the audit state for release '$self'.\n} );

    my $need_audits = 0;

    foreach my $install (@installs) {
        $install->validate_audits || $need_audits++;
    }

    if ( $need_audits ) {
        die qq{Invalid audit state for release '$self'.\n};
    }

    return 1;

}

sub validate_checkpoint {

    my $self = shift;

    my $invalid = 0;
    my $valid   = 0;

    my @installs = $self->installs;

    if ( not @installs ) {
        die qq{The release '$self' has no installs at all.\n};
    }

    STDOUT->print( qq{Validating the checkpoint state for release '$self'.\n} );

    my @validate  = ();
    my %diststate = ();

    foreach my $install (@installs) {

        my $diststate = $diststate{ $install->name } = $install->get_diststate;

        if ( $install->is_empty and not $diststate->get_checktime ) {
            next;
        }

        push @validate, $install;

    }

    if ( not @validate ) {
        die qq{The release '$self' has no installs ready to distribute.\n};
    }

    my $fm_installs = EFS::ForkManager->new({

        list => \@validate,

        parallel => $#validate + 1,

        action => sub {
            my $install = shift;
            $install->validate_checkpoint( diststate => $diststate{ $install->name } );
        },

    });

    $fm_installs->run || die(
        qq{The following installs have invalid checktime states:\n\t},
        join ( qq{\n\t}, sort $fm_installs->get_failed ), qq{\n},
    );

    return 1;

}

sub validate_prod_dependencies {

    my $self = shift;

    my @dependencies = $self->dependencies( type => q{runtime} );

    my @invalid = grep { not $_->target->has_only_prod_stages } @dependencies;

    if ( @invalid ) {
        die(
            qq{The following dependencies are not prod or deprecated:\n\t},
            join ( qq{\n\t}, map { $_->dependency_object_id } @invalid ), qq{\n}
        );
    }

    return 1;

}

sub reverted {
    return shift->get_attribute( name => q{reverted} );
}

1;

# Local Variables:
#   mode: cperl
#   cperl-indent-level: 4
#   fill-column: 78
# End:
# vim: expandtab shiftwidth=4:
