package EFS::Rsync;
use strict;
use warnings;
use English;

use Moose;
use Params::Validate qw( :all );
use Regexp::Common;
use File::Basename;
use IO::Pipe;
use Date::Format;
use Sys::Hostname;
use Readonly;

has q{install} => (
    is  => q{ro},
    isa => q{EFS::DBI::Result::Install},
);

has [qw( src_path dest_path )] => (
    is       => q{ro},
    isa      => q{Str},
    required => 1,
);

has q{src_path_isdir} => (
    is       => q{ro},
    isa      => q{Bool},
    init_arg => undef,
    lazy     => 1,
    default  => sub {
        my $self = shift;
        return -d $self->src_path;
    },
);

has q{dest_path_show} => (
    is       => q{ro},
    isa      => q{Str},
    init_arg => undef,
    lazy     => 1,
    default  => sub {
        my $self = shift;
        return EFS::DBI->admin_path_show($self->dest_path);
    },
);

has [qw( src_cell dest_cell )] => (
    is  => q{ro},
    isa => q{EFS::DBI::Result::Cell},
);

has q{dest_efsserver} => (
    is  => q{rw},
    isa => q{EFS::DBI::Result::EFSServer},
);

has q{query} => (
    is  => q{ro},
    isa => q{Bool},
);

has q{verbose} => (
    is   => q{ro},
    isa  => q{Bool},
    lazy => 1,
    # Query implies verbose
    default => sub {
        my $self = shift;
        return $self->query;
    },
);

has q{diststats_id} => (
    is  => q{ro},
    isa => q{Str},
);

has [qw( bytes_sent bytes_received bytes_rate total_size
         speedup file_count file_xfer_count
         file_size file_xfer_size )] => (
             is       => q{rw},
             isa      => q{Num},
             init_arg => undef,
             default  => 0,
         );

has q{rsyncstats} => (
    is       => q{rw},
    isa      => q{EFS::DBI::Result::RsyncStats},
    init_arg => undef,
);

has q{elapsed_time} => (
    is       => q{rw},
    isa      => q{Int},
    init_arg => undef,
);

has q{exclude} => (
    is  => q{ro},
    isa => q{Str},
);

Readonly my $MAX_RETRIES => 3;

has q{retries} => (
    is      => q{ro},
    isa     => q{Int},
    default => $MAX_RETRIES,
);

has [qw( _deleted _skipped _updated )] => (
    is       => q{ro},
    isa      => q{HashRef[Int]},
    init_arg => undef,
    default  => sub { {} },
);

{

    sub BUILD {

        my ($self,$args_ref) = @_;

        # remove args handled by Moose, as they've already been checked
        foreach my $attr ( qw(install src_path dest_path
                          src_cell dest_cell dest_efsserver
                          query verbose diststats_id
                          exclude retries ) ) {
            delete $args_ref->{$attr};
        }

        my @args = %{ $args_ref };

        my (%args) = validate( @args, {
            dest_efsserver_name => { type => SCALAR, optional => 1 },
        });

        if ( $args{dest_efsserver_name} ) {
            my $efsserver_rs = EFS::DBI->get_resultset( q{EFSServer} );
            my $efsserver = $efsserver_rs->retrieve_bynames_assert(
                name => $args{dest_efsserver_name},
            );
            $self->dest_efsserver( $efsserver );
        }

        $self->run;

        return;

    }

    sub run {

        my $self  = shift;

        my $install = $self->install;

        my $starttime = time;

        my $rsyncstats;

        if ( $self->diststats_id ) {

            my $rsyncstats_rs = EFS::DBI->get_resultset( q{RsyncStats} );

            $rsyncstats = $rsyncstats_rs->create({
                diststats => $self->diststats_id,
                metaproj  => $install->metaproj->name,
                project   => $install->project->name,
                release   => $install->release->name,
                install   => $install->name,
                src_cell  => $self->src_cell->name,
                dest_cell => $self->dest_cell->name,
            });

        }

        my $src_path  = $self->src_path;

        if ( -l $src_path ) {
            die(
                qq{Source path is a symlink: $src_path\n},
                qq{Must be a directory\n},
            );
        }

        if ( $self->src_path_isdir ) {
            $src_path .= q{/};
        }

        my $dest_path = $self->dest_path;

        my $dest_path_info;

        if ( $self->dest_cell ) {
            $dest_path_info = EFS::RPC->execute(
                cell   => $self->dest_cell,
                class  => q{EFS::Util},
                method => q{GetPathInfo},
                args   => [ $dest_path ],
            );
        }
        else {
            $dest_path_info = EFS::Util->GetPathInfo( $dest_path );
        }

        my $dest_path_temp = $dest_path;

        if ( $dest_path_info->{type} eq q{link} ) {
            $dest_path_temp .= q{.new};
        }

        my $pipe = IO::Pipe->new ||
            die qq{Unable to create pipe: $ERRNO\n};

        my $pid = fork;

        if ( not defined $pid ) {
            die qq{Unable to fork: $ERRNO\n};
        }

        my @unknown;

        if ( $pid == 0 ) {

            EFS::DBI->inactivate_handles;

            STDOUT->fdopen( $pipe->writer->fileno, q{w} ) ||
                die qq{Unable to redirect stdout: $ERRNO\n};

            STDERR->fdopen( STDOUT->fileno, q{w} ) ||
                die qq{Unable to redirect stderr: $ERRNO\n};

            STDOUT->autoflush(1);
            STDERR->autoflush(1);

            my @rsync = qw( rsync --numeric-ids --delete
                            --verbose --stats --archive );

            my $efsserver;

            my $rsh_ssh = q{};

            if ( $self->dest_efsserver || $self->dest_cell ) {

                $efsserver =
                    $self->dest_efsserver ||
                        $self->dest_cell->find_efsserver ||
                            die qq{Unable to determine efsserver\n};

                my $ssh = q{ssh -o StrictHostKeyChecking=no -o LogLevel=ERROR};

                my $identity = EFS::CLI::EFSServer->sshid;

                if ( -f $identity ) {
                    $ssh .= qq{ -i $identity};
                }

                my $parameter_rs = EFS::DBI->get_resultset( q{Parameter} );
                if ( my $ssh_port = $parameter_rs->configured( q{ssh_port} ) ) {
                    $ssh .= qq{ -p $ssh_port};
                }

                $rsh_ssh = qq{--rsh=$ssh};

            }

            if ( $self->query ) {
                push @rsync, q{--dry-run};
            }

            if ( $self->exclude ) {
                push @rsync, q{--exclude}, $self->exclude;
            }

            if ( $efsserver and not $efsserver->is_localhost ) {
                push @rsync, $rsh_ssh, $src_path, qq{$efsserver:$dest_path_temp};
            }
            else {
                push @rsync, $src_path, $dest_path_temp;
            }

            if ( $PERLDB || $ENV{EFS_RSYNC_DEBUG} ) {
                my $dump = Data::Dumper->Dump([\@rsync],[q{rsync}]);
                $dump =~ s{^}{###}gmsx;
                print $dump;
            }

            exec (@rsync) ||
                die qq{Unable to exec rsync: $ERRNO\n};

        }
        else {

            $pipe->reader;

        LINE: while ( defined ( my $line = $pipe->getline ) ) {

                if ( $PERLDB  ) {
                    print qq{[raw rsync] $line};
                }

                next if $line =~ m{^\#{3}}msx;

                chomp ($line);
                $line = $RE{ws}{crop}->subs($line);

                next LINE unless $line;

                next LINE if $line eq q{done};
                next LINE if $line =~ m{sending \s+ incremental \s+ file}msx;
                next LINE if $line =~ m/\A building \h file \h list \h \.\.\. \h done/msx;
                next LINE if $line =~ m/\A delta [^\n] transmission/msx;
                next LINE if $line =~ m/\A deleting \h in/msx;
                next LINE if $line =~ m/\[ (sender | receiver) \]/msx;
                next LINE if $line =~ m/total: \h matches/msx;
                next LINE if $line =~ m/\A (Literal | Matched)/msx;
                next LINE if $line =~ m/File \h list \h /msx;

                # We get this data from the verbose summary
                next LINE if $line =~ m/\A Total \s bytes/msx;

                if ( $line =~ m/Number \h of \h files: \h (\d+)/msx ) {
                    $self->file_count($1);
                    next LINE;
                }

                if ( $line
                    =~ m/Number \h of \h files \h transferred: \h (\d+)/msx ) {
                    $self->file_xfer_count($1);
                    next LINE;
                }

                if ( $line =~ m/Total \h file \h size: \h (\d+)/msx ) {
                    $self->file_size($1);
                    next LINE;
                }

                if ( $line
                    =~ m/Total \h transferred \h file \h size: \h (\d+)/msx ) {
                    $self->file_xfer_size($1);
                    next LINE;
                }

                ## no critic ComplexRegexes
                if ($line
                    =~ m{sent \h+ (\d+) \h+ bytes \h+ received \h+ (\d+) \h+
                         bytes \h+ ([\d\.]+) \h+ bytes/sec}msx
                    ) {
                    $self->bytes_sent($1);
                    $self->bytes_received($2);
                    $self->bytes_rate($3);
                    next LINE;
                }

                if ( $line
                    =~ m/total \h size \h is \h (\d+) \h+ speedup \h is \h ([\d\.]+)/msx
                    ) {
                    $self->total_size($1);
                    $self->speedup($2);
                    next LINE;
                }

                if ( $line =~ m/\A deleting (?: \h+ directory)? \h+ ([^\n]*)/msx ) {
                    $self->_deleted->{$1}++;
                    next LINE;
                }

                if ( $line =~ m/\A created \h+ directory \h+ ([^\n]*)/msx ) {
                    $self->_updated->{$1}++;
                    next LINE;
                }

                if ( $line =~ m/([^\n]*) \h is \h uptodate/msx ) {
                    $self->_skipped->{$1}++;
                    next LINE;
                }

                if ( $line =~ m/([^\n]* \s -> \s [^\n]*)/msx ) {
                    $self->_updated->{$1}++;
                    next LINE;
                }

                if ( $line =~ m{([^\n]* /) \z}msx
                    && -d $self->src_path . qq{/$line} ) {
                    $self->_updated->{$1}++;
                    next LINE;
                }

                if ( $self->src_path_isdir ) {
                    if ( -e $self->src_path . qq{/$line} ) {
                        $self->_updated->{$line}++;
                        next LINE;
                    }
                }
                else {
                    if ( $self->src_path =~ m{/$line\z}msx ) {
                        $self->_updated->{$line}++;
                        next LINE;
                    }
                }

                push @unknown, $line;

            }

        }

        my $status;
        my $fatal_message;
        if ( waitpid( $pid, 0 ) ) {
            if ( $status = $CHILD_ERROR ) {
                $fatal_message = qq{Error executing rsync process (exit = $CHILD_ERROR)\n};
            }
        }
        else {
            ##  no critic (MagicNumbers)
            $status        = -1;
            $fatal_message = qq{Unable to get status of rsync process ($pid)};
        }

        if (@unknown) {
            $status = 1;
            $fatal_message = join( qq{\n}, q{Unrecognized rsync output:}, @unknown );
        }

        my $endtime = time;

        if ($rsyncstats) {

            $rsyncstats->bytes_sent( $self->bytes_sent );
            $rsyncstats->bytes_received( $self->bytes_received );
            $rsyncstats->bytes_rate( $self->bytes_rate );
            $rsyncstats->total_size( $self->total_size );
            $rsyncstats->speedup( $self->speedup );
            $rsyncstats->file_count( $self->file_count );
            $rsyncstats->file_xfer_count( $self->file_xfer_count );
            $rsyncstats->file_size( $self->file_size );
            $rsyncstats->file_xfer_size( $self->file_xfer_size );

            $rsyncstats->status($status);

            $rsyncstats->update;

        }

        if ($fatal_message) {
            die $fatal_message;
        }

        if ( $dest_path_info->{type} eq q{link} ) {
            EFS::RPC->execute(
                cell   => $self->dest_cell,
                class  => q{EFS::Util},
                method => q{RenameDir},
                args   => [
                    from => $dest_path_temp,
                    to   => $dest_path,
                ],
            );
        }

        $self->rsyncstats($rsyncstats) if $rsyncstats;

        $self->elapsed_time($endtime - $starttime);

        $self->display_stats if $self->verbose;

        return 1;

    }

    sub display_stats {

        my $self  = shift;

        my $stats   = $self->rsyncstats;
        my $install = $self->install;

        my $summary = qq{Distribution completed:\n}
            . q{Target Path = } . $self->dest_path_show . qq{\n};

        my $transfered  =
            ( $self->query ? q{Need to transfer} : q{Transfered} );
        my $paths       =
            ( $self->query ? q{Paths that need to be} : q{Paths} );

        $summary .= sprintf (
            qq{$transfered %d of %d total files, %d of %d total bytes\n},
            $self->file_xfer_count, $self->file_count,
            $self->file_xfer_size,  $self->file_size
        );

        if ( %{ $self->_updated } ) {
            $summary .= qq{$paths updated:\n\t}
                . join ( qq{\n\t}, $self->get_updated ) . qq{\n};
        }

        if ( %{ $self->_deleted } ) {
            $summary .= qq{$paths deleted:\n\t}
                . join ( qq{\n\t}, $self->get_deleted ) . qq{\n};
        }

        $summary .= q{Elapsed time = } . $self->elapsed_time . qq{ secs\n};

        return STDOUT->print($summary);

    }

    sub get_updated {
        my $self  = shift;
        my @updated = ();
        foreach my $path ( sort keys %{ $self->_updated } ) {
            if ( $path ne q{./} and $path ne $self->dest_path ) {
                push @updated, $path;
            }
        }
        return @updated;
    }

    sub get_deleted {
        my $self  = shift;
        return @{ [ sort keys %{ $self->_deleted } ] };
    }

}

no Moose;
__PACKAGE__->meta->make_immutable;
1;

# Local Variables:
#   mode: cperl
#   cperl-indent-level: 4
#   fill-column: 78
# End:
# vim: expandtab shiftwidth=4:
