#!/usr/bin/env perl
use strict;
use warnings;
use English;

use Config;

use blib;
use EFS;

use lib q{test/lib};
use EFS::Test;
use EFS::Test::Config;

use Test::More tests => 11;
use IO::File;
use IO::Socket::SSL;
use Socket;
use File::Path qw(rmtree);

my $config        = EFS::Test::Config->new;
my $ldc_efsserver = $config->get_value( q{LocalDevCell}, q{efsserver} );
my $rdc_efsserver = $config->get_value( q{RemoteDevCell}, q{efsserver} );
my $scpcmd        = $config->get_value( q{Global}, q{sshcmd} );
$scpcmd           =~ s/^ssh/scp/msx;

SKIP: {

    if ( not $config->get_value( q{Security}, q{kerberos_realm} ) ) {
        skip q{Kerberos tests not configured}, 1;
    }

    ## no critic: RequireInterpolationOfMetachars
    my $krb5ccname = qq{/var/tmp/krb5_efstest_$PID};
    $ENV{KRB5CCNAME} = qq{FILE:$krb5ccname};
    system qq{kinit -k -t /var/kerberos/keytabs/efsops.keytab efsops};
    ok( $CHILD_ERROR == 0, q{Acquired efsops credentials for this test} );

}

my $efsserver_port = EFS::CLI->get_serverport;

my $ssl_tmpdir = q{/var/tmp/.efsd.ssl};

EFS::Util->MakeDir( path => $ssl_tmpdir, mode => 0700 );
ok( -d $ssl_tmpdir, q{Created temporary ssl directory} );

# Copy the existing certs for this host, and the remote efsserver

my $ssl_keyfile_local
    = EFS::DBI::Result::EFSServer->get_ssl_keyfile($ldc_efsserver);
my $ssl_certfile_local
    = EFS::DBI::Result::EFSServer->get_ssl_certfile($ldc_efsserver);

system (qq{cp $ssl_keyfile_local $ssl_certfile_local $ssl_tmpdir});
ok( $CHILD_ERROR == 0, q{Copied local cert and keys} );

my $ssl_keyfile_remote
    = EFS::DBI::Result::EFSServer->get_ssl_keyfile($rdc_efsserver);
my $ssl_certfile_remote
    = EFS::DBI::Result::EFSServer->get_ssl_certfile($rdc_efsserver);

my $sshid = EFS::CLI::EFSServer->sshid;

system (qq{$scpcmd -i $sshid $rdc_efsserver:$ssl_keyfile_remote $ssl_tmpdir});
ok( $CHILD_ERROR == 0, q{Copied remote cert and keys} );

# Make sure we can connect using the valid certs

my $valid = IO::Socket::SSL->new(
    PeerAddr      => $rdc_efsserver,
    PeerPort      => $efsserver_port,
    Proto         => q{tcp},
    Type          => SOCK_STREAM,
    Timeout       => 60,
    SSL_use_cert  => 1,
    SSL_version   => q{SSLv3},
    SSL_key_file  => $ssl_keyfile_local,
    SSL_cert_file => $ssl_certfile_local,
);
ok( ref $valid && $valid->isa(q{IO::Socket::SSL}),
    q{Able to connect using the valid certs}
);

ok( $valid->close, q{Closed the temporary connection} );

# Make sure the remote ones don't work

my $stolen = IO::Socket::SSL->new(
    PeerAddr      => $rdc_efsserver,
    PeerPort      => $efsserver_port,
    Proto         => q{tcp},
    Type          => SOCK_STREAM,
    Timeout       => 60,
    SSL_use_cert  => 1,
    SSL_version   => q{SSLv3},
    SSL_key_file  => qq{$ssl_tmpdir/$rdc_efsserver-key.pem},
    SSL_cert_file => qq{$ssl_tmpdir/$rdc_efsserver-cert.pem},
);
ok( not ( ref $stolen && $stolen->isa(q{IO::Socket::SSL}) ),
    q{Unable to connect to using the stolen certs}
);

# Generate new ones, and revoke the old ones

ok( EFS::CLI::EFSServer->newcerts,   q{Generated new SSL certificates} );
ok( EFS::CLI::EFSServer->purgecerts, q{Purged old SSL certificates} );

# Make sure the old ones don't work

my $invalid = IO::Socket::SSL->new(
    PeerAddr      => $rdc_efsserver,
    PeerPort      => $efsserver_port,
    Proto         => q{tcp},
    Type          => SOCK_STREAM,
    Timeout       => 60,
    SSL_use_cert  => 1,
    SSL_version   => q{SSLv3},
    SSL_key_file  => qq{$ssl_tmpdir/$ldc_efsserver-key.pem},
    SSL_cert_file => qq{$ssl_tmpdir/$ldc_efsserver-cert.pem},
);
ok( not ( ref $invalid && $invalid->isa(q{IO::Socket::SSL}) ),
    q{Unable to connect to using the revoked certs}
);

ok( rmtree($ssl_tmpdir), q{Purged temporary ssl directory} );

END {
    my $krb5ccname = qq{/var/tmp/krb5_efstest_$PID};
    if ( $krb5ccname && -f $krb5ccname ) {
        system q{kdestroy};
        $CHILD_ERROR and warn qq{Unable to destroy kerberos credentials\n};
    }
}

