#!/usr/bin/env perl

use strict;
use warnings;
use English;

use Test::More;

use lib q{test/lib};
use EFS::Test;

my $test = EFS::Test->new;
my $config = $test->get_config;

my $username  = $test->get_primary_username;
my $groupname = $test->get_primary_groupname;

my $ldc_name    = $config->get_value( q{LocalDevCell}, q{name} );
my $rdc_name    = $config->get_value( q{RemoteDevCell}, q{name} );
my $lpc_name    = $config->get_value( q{LocalProdCell}, q{name} );
my $root_domain = $config->get_value( q{Global}, q{root_domain} );

my $ownerlist = join q{@}, $username, $root_domain;

$test->efs_test_command(
    command => q{efs display metaproj testmeta},
    not     => 1,
    bail    => 1,
    expect  => q{No such metaproj 'testmeta'},
);

$test->efs_test_command(
    run_as  => q{efsops},
    command => qq{efs define metaproj testmeta efsops efsops $ownerlist},
    not     => 1,
    expect  => [
        q{'efs define metaproj' has been deprecated},
        q{Use 'efs create metaproj' instead},
    ],
);

$test->efs_test_command(
    run_as  => q{efsops},
    command => qq{efs create metaproj testmeta nosuchus $groupname $ownerlist},
    not     => 1,
    expect  => q{No such username 'nosuchus'},
);

$test->efs_test_command(
    run_as  => q{efsops},
    command => qq{efs create metaproj testmeta $username nosuchgr $ownerlist},
    not     => 1,
    expect  => q{No such groupname 'nosuchgr'},
);

$test->efs_test_command(
    run_as  => q{efsops},
    command => qq{efs create metaproj testmeta efsops efsops $ownerlist},
    expect => [
        q{Created metaproj 'testmeta' with username 'efsops', } .
            qq{groupname 'efsops', owner_list '$ownerlist'},
    ],
);

$test->efs_test_command(
    run_as  => q{efsops},
    command => qq{efs create metaproj testmeta $username $groupname $ownerlist},
    expect => [
        qq{Updated metaproj 'testmeta', set username to '$username'},
        qq{Updated metaproj 'testmeta', set groupname to '$groupname'},
    ],
);

$test->efs_test_command(
    command => q{efs display metaproj testmeta},
    expect  => [
        qr{Metaproj\s+Username\s+Groupname\s+Spaceused\s+Filecount\s+Dircount\s+ Linkcount}msx,
        qr{^=+}msx,
        qr{testmeta \s+ $username \s+ $groupname \s+ 0 \s+ bytes \s+ 0 \s+ 0 \s+ 0}msx,
    ],
);

$test->efs_test_command(
    run_as  => q{efsops},
    command => qq{efs define userrights testmeta admin $username},
    expect  => qq{Granted $username admin rights to testmeta},
);

$test->efs_test_command(
    command => qq{efs define userrights testmeta prod $username},
    expect  => qq{Granted $username prod rights to testmeta},
);

$test->efs_test_command(
    command => qq{efs create project nosuch testproj $ldc_name},
    not     => 1,
    expect  => [
        q{No such metaproj 'nosuch'},
        q{Exit status from efs server not received.  Assuming failure...},
    ],
);

$test->efs_test_command(
    command => q{efs create project testmeta testproj no.such},
    not     => 1,
    expect  => q{No cell name matching 'no.such*'},
);

$test->efs_test_command(
    command => qq{efs create project testmeta testproj $lpc_name},
    not     => 1,
    expect  => qq{Unable to create projects in non-dev cell '$lpc_name'},
);


$test->efs_test_command(
    # NOTE: Some of the directory messages may or may not be there.
    # If you rebuild everything from scratch, they are.
    # If you rerun 20_efsops, they aren't.
    filter =>
        qr{Created \s+ ( project | symlink )}msx,
    command   => qq{efs create project testmeta testproj $ldc_name},
    expect    => [
        qr{Created \s+ symlink \s+ $ldc_name:/efs/dev/testmeta/testproj}msx,
        qr{Created \s+ symlink \s+ $rdc_name:/efs/dev/testmeta/testproj}msx,
        q{Created project testmeta/testproj},
    ],
);

$test->efs_test_command(
    command => q{efs display project testmeta testproj},
    expect  => [
        qr{Metaproj\s+Project\s+Cell\s+Spaceused\s+Filecount\s+Dircount\s+Linkcount}msx,
        qr{^=+}msx,
        qr{testmeta \s+ testproj \s+ $ldc_name \s+ 0 \s+ bytes \s+ 0 \s+ 0 \s+ 0}msx,
    ],
);

$test->efs_test_command(
    command => q{efs display metaproj testmeta},
    expect  => [
        qr{Metaproj\s+Username\s+Groupname\s+Spaceused\s+Filecount\s+Dircount\s+ Linkcount}msx,
        qr{^=+}msx,
        qr{testmeta \s+ $username \s+ $groupname \s+ 0 \s+ bytes \s+ 0 \s+ 0 \s+ 0}msx,
        qr{Metaproj \s+ Role \s+ Username}msx,
        qr{^=+}msx,
        qr{testmeta \s+ admin \s+ $username}msx,
        qr{testmeta \s+ prod \s+ $username}msx,
        qr{Metaproj\s+Project\s+Cell\s+Spaceused\s+Filecount\s+Dircount\s+Linkcount}msx,
        qr{^=+}msx,
        qr{testmeta \s+ testproj \s+ $ldc_name \s+ 0 \s+ bytes \s+ 0 \s+ 0 \s+ 0}msx,
    ],
);

$test->efs_test_command(
    command => qq{efs create project testmeta testproj $rdc_name},
    not     => 1,
    expect  =>
        qq{Project 'testmeta/testproj' already exists in cell '$ldc_name'},
);

$test->efs_test_command(
    run_as  => q{efsops},
    command => q{efs destroy metaproj testmeta},
    not     => 1,
    expect  => [
        q{Unable to destroy metaproj 'testmeta'},
        q{The following projects still exist:},
        q{testmeta/testproj},
    ],
);

# NOTE: We need a separate set of tests for deprecated functions.
# This doesn't belong here, but this is as good as any place, really.

$test->efs_test_command(
    run_as  => q{efsops},
    command => q{efs set_dist2dev project testmeta testproj},
    expect  => q{The set_dist2dev command has been deprecated},
);

done_testing();

exit 0;
